import click
from arbol.arbol import aprint, asection

from dexp.cli.dexp_main import _default_clevel, _default_codec, _default_store
from dexp.cli.dexp_main import _default_workers_backend
from dexp.cli.parsing import _parse_channels, _get_output_path, _parse_slicing
from dexp.datasets.open_dataset import glob_datasets
from dexp.datasets.operations.copy import dataset_copy


@click.command()
@click.argument('input_paths', nargs=-1)  # ,  help='input path'
@click.option('--output_path', '-o')  # , help='output path'
@click.option('--channels', '-c', default=None, help='List of channels, all channels when ommited.')
@click.option('--slicing', '-s', default=None, help='Dataset slice (TZYX), e.g. [0:5] (first five stacks) [:,0:100] (cropping in z) ')
@click.option('--store', '-st', default=_default_store, help='Zarr store: ‘dir’, ‘ndir’, or ‘zip’', show_default=True)
@click.option('--codec', '-z', default=_default_codec, help='Compression codec: zstd for ’, ‘blosclz’, ‘lz4’, ‘lz4hc’, ‘zlib’ or ‘snappy’ ', show_default=True)
@click.option('--clevel', '-l', type=int, default=_default_clevel, help='Compression level', show_default=True)
@click.option('--overwrite', '-w', is_flag=True, help='Forces overwrite of target', show_default=True)
@click.option('--zerolevel', '-zl', type=int, default=0, help="‘zero-level’ i.e. the pixel values in the restoration (to be substracted)", show_default=True)  #
@click.option('--workers', '-wk', default=-4, help='Number of worker threads to spawn. Negative numbers n correspond to: number_of _cores / |n| ', show_default=True)  #
@click.option('--workersbackend', '-wkb', type=str, default=_default_workers_backend, help='What backend to spawn workers with, can be ‘loky’ (multi-process) or ‘threading’ (multi-thread) ', show_default=True)  #
@click.option('--check', '-ck', default=True, help='Checking integrity of written file.', show_default=True)  #
def copy(input_paths,
         output_path,
         channels,
         slicing,
         store,
         codec,
         clevel,
         overwrite,
         zerolevel,
         workers,
         workersbackend,
         check):
    """ Copies a dataset, channels can be selected, cropping can be performed, compression can be changed, ...
    """

    input_dataset, input_paths = glob_datasets(input_paths)
    output_path = _get_output_path(input_paths[0], output_path, '_copy')
    slicing = _parse_slicing(slicing)
    channels = _parse_channels(input_dataset, channels)

    with asection(f"Copying from: {input_paths} to {output_path} for channels: {channels}, slicing: {slicing} "):
        dataset_copy(input_dataset,
                     output_path,
                     channels=channels,
                     slicing=slicing,
                     store=store,
                     compression=codec,
                     compression_level=clevel,
                     overwrite=overwrite,
                     zerolevel=zerolevel,
                     workers=workers,
                     workersbackend=workersbackend,
                     check=check)

        input_dataset.close()
        aprint("Done!")



import math
import os
from os.path import exists, join

import click
import numpy
from arbol.arbol import section, aprint, asection

from dexp.cli.parsing import _parse_channels, _parse_slicing
from dexp.datasets.open_dataset import glob_datasets


@click.command()
@click.argument('input_paths', nargs=-1)
@click.option('--output_path', '-o', default=None, help='Output folder to store rendered PNGs. Default is: frames_<channel_name>')
@click.option('--channels', '-c', default=None, help='list of channels to color, all channels when ommited.')
@click.option('--slicing', '-s', default=None, help='dataset slice (TZYX), e.g. [0:5] (first five stacks) [:,0:100] (cropping in z).')
@click.option('--overwrite', '-w', is_flag=True, help='to force overwrite of target', show_default=True)  # , help='dataset slice'
@click.option('--aspect', '-a', type=float, default=4, help='sets aspect ratio e.g. 4', show_default=True)
@click.option('--colormap', '-cm', type=str, default='magma', help='sets colormap, e.g. viridis, gray, magma, plasma, inferno ', show_default=True)
@click.option('--rendersize', '-rs', type=int, default=2048, help='Sets the frame color size. i.e. -ws 400 sets the window to 400x400', show_default=True)
@click.option('--clim', '-cl', type=str, default='0,1024', help='Sets the contrast limits, i.e. -cl 0,1000 sets the contrast limits to [0,1000]')
@click.option('--options', '-opt', type=str, default=None, help='Render options, e.g: \'gamma=1.2,box=True\'. Important: no white spaces!!! Complete list with defaults will be displayed on first run')
def volrender(input_paths, output_path, channels=None, slicing=None, overwrite=False, aspect=None, colormap='viridis', rendersize=1536, clim=None, options=None):
    """ Renders 3D volumetric video
    """

    input_dataset, input_paths = glob_datasets(input_paths)
    channels = _parse_channels(input_dataset, channels)
    slicing = _parse_slicing(slicing)

    aprint(f"Volumetric rendering of: {input_paths} to {output_path} for channels: {channels}, slicing: {slicing} ")

    for channel in channels:

        with asection(f"Channel '{channel}' shape: {input_dataset.shape(channel)}:"):
            aprint(input_dataset.info(channel))

        array = input_dataset.get_array(channel, wrap_with_dask=True)
        dtype = array.dtype

        if slicing:
            array = array[slicing]

        aprint(f"Rendering array of shape={array.shape} and dtype={array.dtype} for channel '{channel}'.")

        if clim is None:
            aprint(f"Computing min and max from first stack...")
            first_stack = numpy.array(input_dataset.get_stack(channel, 0, per_z_slice=False))
            min_value = max(0, first_stack.min() - 100)
            max_value = first_stack.max() + 100
            aprint(f"min={min_value} and max={max_value}.")
        else:
            aprint(f"provided min and max for contrast limits: {clim}")
            min_value, max_value = (float(strvalue) for strvalue in clim.split(','))

        aprint(f"Provided rendering options: {options}")
        options = dict(item.split("=") for item in options.split(",")) if options is not None else dict()

        def str2bool(v):
            return v.lower() in ("yes", "true", "t", "1")

        nbtp = array.shape[0]
        nbframes = int(options['nbframes']) if 'nbframes' in options else 1010

        skip = int(options['skip']) if 'skip' in options else 1
        cut = str(options['cut']) if 'cut' in options else 'nocut'
        cutpos = float(options['cutpos']) if 'cutpos' in options else 0
        cutspeed = float(options['cutspeed']) if 'cutspeed' in options else 0
        irot = options['irot'] if 'irot' in options else ''
        raxis = options['raxis'] if 'raxis' in options else 'y'
        rstart = float(options['rstart']) if 'rstart' in options else 0
        rspeed = float(options['rspeed']) if 'rspeed' in options else 0.15
        tstart = int(options['tstart']) if 'tstart' in options else 0
        tspeed = float(options['tspeed']) if 'tspeed' in options else 0.5
        gamma = float(options['gamma']) if 'gamma' in options else 0.9
        zoom = float(options['zoom']) if 'zoom' in options else 1.45
        alpha = float(options['alpha']) if 'alpha' in options else 0.3
        box = str2bool(options['box']) if 'box' in options else False
        maxsteps = int(options['maxsteps']) if 'maxsteps' in options else 1024
        norm = str2bool(options['norm']) if 'norm' in options else True
        normrange = float(options['normrange']) if 'normrange' in options else 1024
        videofilename = options['name'] if 'name' in options else 'video.mp4'

        with asection("Options:"):
            aprint(f"Video filename          : {videofilename}")
            aprint(f"Number of time points   : {nbtp}")
            aprint(f"Number of frames        : {nbframes}  \toption: nbframes: \tint")
            aprint(f"Render one frame every  : {skip}      \toption: skip:     \tint")
            aprint(f"Volume cutting          : {cut}       \toption: cut:      \t[left, right, top, bottom, front, back, none]")
            aprint(f"Volume cutting position : {cutpos}    \toption: cutpos:   \tfloat")
            aprint(f"Volume cutting speed    : {cutspeed}  \toption: cutspeed: \tfloat")
            aprint(f"Initial time point      : {tstart}    \toption: tstart:   \tint")
            aprint(f"Time    speed           : {tspeed}    \toption: tspeed:   \tfloat")
            aprint(f"Initial rotation        : {irot}      \toption: irot:     \t xxyzzz -> 90deg along x, 45deg along y, 135deg along z ")
            aprint(f"Rotation axis           : {raxis}     \toption: raxis:    \t[x,y,z]")
            aprint(f"Initial rotation angle  : {rstart}    \toption: rstart:   \tfloat")
            aprint(f"Rotation speed          : {rspeed}    \toption: rspeed:   \tfloat")
            aprint(f"Gamma                   : {gamma}     \toption: gamma:    \tfloat")
            aprint(f"Zoom                    : {zoom}      \toption: zoom:     \tfloat")
            aprint(f"Alpha                   : {alpha}     \toption: alpha:    \tfloat")
            aprint(f"box?                    : {box}       \toption: box:      \tbool (true/false)")
            aprint(f"normalisation           : {norm}      \toption: norm:     \tbool (true/false)")
            aprint(f"normalisation range     : {normrange} \toption: normrange:\tfloat")
            aprint(f"Max steps for vol color: {maxsteps}  \toption: maxsteps: \tint")

        if output_path is None:
            output_path = f"frames_{channel}"
        os.makedirs(output_path, exist_ok=True)
        from spimagine import volshow
        from spimagine import DataModel
        from spimagine import NumpyData

        aprint("Opening spimagine...")
        import spimagine
        spimagine.config.__DEFAULTMAXSTEPS__ = maxsteps
        spimagine.config.__DEFAULT_TEXTURE_WIDTH__ = rendersize

        datamodel = DataModel(NumpyData(array[0].compute()))
        aprint("Creating Spimagine window... (you can minimise but don't close!)")
        win = volshow(datamodel, stackUnits=(1., 1., aspect), autoscale=False, show_window=True)
        win.resize(rendersize + 32, rendersize + 32)
        win.showMinimized()

        while section("Rendering:"):
            for i in range(0, nbframes, skip):
                aprint(f"______________________________________________________________________________")
                aprint(f"Frame     : {i}")

                tp = tstart + int(tspeed * i)
                if tp >= nbtp:
                    break
                aprint(f"Time point: {tp}")

                angle = rstart + rspeed * i
                aprint(f"Angle     : {angle}")

                effcutpos = cutpos + cutspeed

                filename = join(output_path, f"frame_{i:05}.png")

                if overwrite or not exists(filename):

                    aprint("Loading stack...")
                    stack = array[int(tp)].compute()

                    if norm:
                        aprint("Computing percentile...")
                        rmax = numpy.percentile(stack[::8].astype(numpy.float32), q=99.99).astype(numpy.float32)
                        # rmax = numpy.max(stack[::8]).astype(numpy.float32)
                        aprint(f"rmax={rmax}")

                        aprint("Normalising...")
                        norm_max_value = normrange
                        norm_min_value = 64.0
                        # stack = norm_min_value+stack*((norm_max_value-norm_min_value)/rmax)
                        stack = stack * numpy.array((norm_max_value - norm_min_value) / rmax, dtype=numpy.float32)
                        stack += numpy.array(norm_min_value, dtype=dtype)
                        stack = stack.astype(dtype)

                    # print("Opening spimagine...")
                    # win = volshow(stack, stackUnits=(1., 1., aspect), autoscale=False, show_window=True)

                    aprint("Loading stack into Spimagine...")
                    datamodel.setContainer(NumpyData(stack))
                    win.setModel(datamodel)

                    aprint("Setting rendering parameters...")
                    if colormap in spimagine.config.__COLORMAPDICT__:
                        win.set_colormap(colormap)
                    else:
                        from matplotlib import colors
                        rgb = colors.to_rgba(colormap)[:3]
                        aprint(f"Turning the provided color: {colormap} = {rgb} into a colormap.")
                        win.set_colormap_rgb(rgb)

                    win.transform.setStackUnits(1., 1., aspect)
                    win.transform.setGamma(gamma)
                    win.transform.setMin(min_value)
                    win.transform.setMax(max_value)
                    win.transform.setZoom(zoom)
                    win.transform.setAlphaPow(alpha)
                    win.transform.setBox(box)

                    if cut == 'left':
                        win.transform.setBounds(effcutpos, 1, -1, 1, -1, 1)
                    elif cut == 'right':
                        win.transform.setBounds(-1, effcutpos, -1, 1, -1, 1)
                    elif cut == 'top':
                        win.transform.setBounds(-1, 1, effcutpos, 1, -1, 1)
                    elif cut == 'bottom':
                        win.transform.setBounds(-1, 1, -1, effcutpos, -1, 1)
                    elif cut == 'front':
                        win.transform.setBounds(-1, 1, -1, 1, effcutpos, 1)
                    elif cut == 'back':
                        win.transform.setBounds(-1, 1, -1, 1, -1, effcutpos)
                    elif cut == 'centerx':
                        win.transform.setBounds(-0.25 - effcutpos, 0.25 + effcutpos, -1, 1, -1, 1)
                    elif cut == 'centery':
                        win.transform.setBounds(-1, 1, -0.25 - effcutpos, 0.25 + effcutpos, -1, 1)
                    elif cut == 'centerz':
                        win.transform.setBounds(-1, 1, -1, 1, -0.25 - effcutpos, 0.25 + effcutpos)
                    elif cut == 'none':
                        win.transform.setBounds(-1, 1, -1, 1, -1, 1)

                    win.transform.setRotation(0, 1, 0, 0)

                    for character in irot:
                        if character == 'x':
                            aprint(f"Rotating along x axis by 45 deg (prev quatRot={win.transform.quatRot})")
                            win.transform.addRotation(0.5 * math.pi / 4, 1, 0, 0)
                        elif character == 'y':
                            aprint(f"Rotating along y axis by 45 deg (prev quatRot={win.transform.quatRot})")
                            win.transform.addRotation(0.5 * math.pi / 4, 0, 1, 0)
                        elif character == 'z':
                            aprint(f"Rotating along z axis by 45 deg (prev quatRot={win.transform.quatRot})")
                            win.transform.addRotation(0.5 * math.pi / 4, 0, 0, 1)
                    aprint(f"Rotation after initial axis rotation: {win.transform.quatRot}")

                    if 'x' in raxis:
                        aprint(f"Rotating along x axis by {angle} deg (prev quatRot={win.transform.quatRot})")
                        win.transform.addRotation(0.5 * angle * (math.pi / 180), 1, 0, 0)
                    if 'y' in raxis:
                        aprint(f"Rotating along y axis by {angle} deg (prev quatRot={win.transform.quatRot})")
                        win.transform.addRotation(0.5 * angle * (math.pi / 180), 0, 1, 0)
                    if 'z' in raxis:
                        aprint(f"Rotating along z axis by {angle} deg (prev quatRot={win.transform.quatRot})")
                        win.transform.addRotation(0.5 * angle * (math.pi / 180), 0, 0, 1)

                    aprint(f"Final rotation: {win.transform.quatRot}")

                    aprint(f"Saving frame: {filename}")
                    win.saveFrame(filename)

    win.closeMe()
    input_dataset.close()
    aprint("Done!")

    raise SystemExit
    import sys
    sys.exit()



from os.path import join

from arbol import asection, aprint


def perform_speed_test():

    import os
    cwd = os.getcwd()
    with asection(f"Measuring write/read at current path: {cwd}"):
        import subprocess

        filename = 'zz__speedtestfile__zz'
        filepath = join(cwd, filename)

        from sys import platform
        if platform == "linux" or platform == "linux2" or platform == "darwin":
            with asection("Write speed:"):
                result = subprocess.run(['time', 'sh', '-c', f'dd if=/dev/zero of="{filepath}" bs=4194304 count=256 && sync'], capture_output=True)
                # aprint(result.stdout.decode("utf-8").split('\n'))
                aprint(result.stderr.decode("utf-8").split('\n')[2])

            with asection("Read speed:"):
                result = subprocess.run(['time', 'sh', '-c', f'dd if="{filepath}" of=/dev/null bs=4194304 count=256 && sync'], capture_output=True)
                # aprint(result.stdout.decode("utf-8").split('\n'))
                aprint(result.stderr.decode("utf-8").split('\n')[2])

            with asection("Read speed with cache clearing:"):
                # clear cache:
                return_code = os.system('sync; echo 3 > /proc/sys/vm/drop_caches')
                if return_code != 0:
                    aprint("you must be root to clear the cache!")

                result = subprocess.run(['time', 'sh', '-c', f'dd if="{filepath}" of=/dev/null bs=4194304 count=256 && sync'], capture_output=True)
                # aprint(result.stdout.decode("utf-8").split('\n'))
                aprint(result.stderr.decode("utf-8").split('\n')[2])



        elif platform == "win32":
            aprint(f"Not implemented yet for operating system: {platform}")
        else:
            aprint(f"Unknown operating system: {platform}")

        try:
            os.remove(filepath)
        except:
            aprint(f"Could not delete file: {filepath}")


import tempfile
import warnings
from os.path import join

import numpy as np
from csbdeep.data import RawData, create_patches, no_background_patches
from csbdeep.data.transform import anisotropic_distortions
from csbdeep.io import load_training_data
from csbdeep.io import save_training_data
from csbdeep.models import Config, IsotropicCARE


class IsoNet:

    def __init__(self, context_name='default', subsampling=5):
        """

        """

        self.context_folder = join(tempfile.gettempdir(), context_name)
        self.training_data_path = join(self.context_folder, "training_data.npz")
        self.model_path = join(self.context_folder, "model")

        self.model = None

        self.subsampling = subsampling

    def prepare(self, image, psf=np.ones((3, 3)) / 9, threshold=0.9):
        print('image size         =', image.shape)
        print('Z subsample factor =', self.subsampling)

        raw_data = RawData.from_arrays(image, image, axes='ZYX')

        anisotropic_transform = anisotropic_distortions(
            subsample=self.subsampling,
            psf=psf,  # use the actual PSF here
            psf_axes='YX',
        )

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            X, Y, XY_axes = create_patches(
                raw_data=raw_data,
                patch_size=(128, 128),
                n_patches_per_image=512,
                transforms=[anisotropic_transform],
                patch_filter=no_background_patches(threshold=threshold, percentile=100 - 0.1),
            )

        assert X.shape == Y.shape
        print("shape of X,Y =", X.shape)
        print("axes  of X,Y =", XY_axes)

        print("Saving training data...")
        save_training_data(self.training_data_path, X, Y, XY_axes)
        print("Done saving training data.")

    def train(self, max_epochs=100):
        (X, Y), (X_val, Y_val), axes = load_training_data(self.training_data_path, validation_split=0.1, verbose=True)

        n_channel_in, n_channel_out = 1, 1

        config = Config(axes, n_channel_in, n_channel_out, train_steps_per_epoch=30)
        print(config)
        vars(config)

        model = IsotropicCARE(config, 'my_model', basedir=self.model_path)

        history = model.train(X, Y, validation_data=(X_val, Y_val), epochs=max_epochs)

        print(sorted(list(history.history.keys())))

    def apply(self, image, batch_size=8):
        if self.model is None:
            print('Loading model.')
            self.model = IsotropicCARE(config=None, name='my_model', basedir=self.model_path)
        axes = 'ZYX'

        print(f'Applying model to image of shape: {image.shape}...')
        restored = self.model.predict(image, axes, self.subsampling, batch_size=batch_size)
        print('done!')

        return restored



import tempfile
from os.path import join

import numpy
from skimage.data import binary_blobs
from skimage.filters import gaussian

from dexp.datasets.zarr_dataset import ZDataset
from dexp.processing.backends.numpy_backend import NumpyBackend


def test_zarr_dataset_livecycle():
    with tempfile.TemporaryDirectory() as tmpdir:
        print('created temporary directory', tmpdir)

        # self, path:str, mode:str ='r', store:str ='dir'
        zdataset = ZDataset(path=join(tmpdir, 'test.zarr'),
                            mode='w',
                            store='dir')

        array1 = zdataset.add_channel(name='first',
                                      shape=(10, 100, 100, 100),
                                      chunks=(1, 50, 50, 50),
                                      dtype='f4',
                                      codec='zstd',
                                      clevel=3)

        array2 = zdataset.add_channel(name='second',
                                      shape=(17, 10, 20, 30),
                                      chunks=(1, 5, 1, 6),
                                      dtype='f4',
                                      codec='zstd',
                                      clevel=3)

        assert array1 is not None
        assert array2 is not None

        assert len(zdataset.channels()) == 2
        assert 'first' in zdataset.channels()
        assert 'second' in zdataset.channels()

        assert zdataset.shape('first') == (10, 100, 100, 100)
        assert zdataset.chunks('first') == (1, 50, 50, 50)
        assert zdataset.dtype('first') == 'f4'

        assert zdataset.shape('second') == (17, 10, 20, 30)
        assert zdataset.chunks('second') == (1, 5, 1, 6)
        assert zdataset.dtype('second') == 'f4'

        assert not zdataset.check_integrity()

        with NumpyBackend() as backend:
            xp = backend.get_xp_module()
            blobs = binary_blobs(length=100, n_dim=3, blob_size_fraction=0.1).astype('f4')
            for i in range(0, 10):

                blobs = gaussian(blobs, sigma=1 + 0.1 * i)
                zdataset.write_stack('first', i, blobs)

                # import napari
                # with napari.gui_qt():
                #     from napari import Viewer
                #     viewer = Viewer()
                #     viewer.add_image(blobs, name='blobs')
                #     viewer.add_image(zdataset.get_stack('first', i), name='(zdataset.get_stack(first, i)')

                assert xp.all((zdataset.get_stack('first', i) == blobs))

                for axis in range(3):
                    assert xp.all((zdataset.get_projection_array('first', axis=axis)[i] == xp.max(blobs, axis=axis)))

            blobs = binary_blobs(length=30, n_dim=3, blob_size_fraction=0.03).astype('f4')
            for i in range(0, 17):
                blobs = gaussian(blobs, sigma=1 + 0.1 * i)
                blobs = blobs[0:10, 0:20, 0:30]
                zdataset.write_stack('second', i, blobs)

                assert xp.all((zdataset.get_stack('second', i) == blobs))

                for axis in range(3):
                    assert xp.all((zdataset.get_projection_array('second', axis=axis)[i] == xp.max(blobs, axis=axis)))

        assert zdataset.check_integrity()

        # with napari.gui_qt():
        #     viewer = Viewer()
        #     viewer.add_image(array1, name='image')


def test_zarr_integrity():
    with tempfile.TemporaryDirectory() as tmpdir:
        print('created temporary directory', tmpdir)

        # self, path:str, mode:str ='r', store:str ='dir'
        zdataset = ZDataset(path=join(tmpdir, 'test.zarr'),
                            mode='w',
                            store='dir')

        array = zdataset.add_channel(name='first',
                                     shape=(10, 100, 100, 100),
                                     chunks=(1, 50, 50, 50),
                                     dtype='f4',
                                     codec='zstd',
                                     clevel=3)

        # we initialise almost everything:
        array[0:9] = 0

        # the dataset integrity must be False!
        assert not zdataset.check_integrity()

        # we initialise everything:
        array[...] = 1
        # the dataset integrity must be True!
        assert zdataset.check_integrity()


def test_add_channels_to():
    with tempfile.TemporaryDirectory() as tmpdir:
        print('created temporary directory', tmpdir)

        # self, path:str, mode:str ='r', store:str ='dir'
        dataset1_path = join(tmpdir, 'test1.zarr')
        zdataset1 = ZDataset(path=dataset1_path,
                             mode='w',
                             store='dir')

        array1 = zdataset1.add_channel(name='first',
                                       shape=(10, 11, 12, 13),
                                       chunks=None,
                                       dtype='f4',
                                       codec='zstd',
                                       clevel=3)

        array1[...] = 1

        dataset2_path = join(tmpdir, 'test2.zarr')
        zdataset2 = ZDataset(path=dataset2_path,
                             mode='w',
                             store='dir')

        array2 = zdataset2.add_channel(name='second',
                                       shape=(17, 10, 20, 30),
                                       chunks=None,
                                       dtype='f4',
                                       codec='zstd',
                                       clevel=3)

        array2[...] = 1

        zdataset2.add_channels_to(dataset1_path,
                                  channels=('second',),
                                  rename=('second-',),
                                  )

        zdataset1_reloaded = ZDataset(path=dataset1_path,
                                      mode='r',
                                      store='dir')

        assert len(zdataset1_reloaded.channels()) == 2
        assert 'first' in zdataset1_reloaded.channels()
        assert 'second-' in zdataset1_reloaded.channels()

        assert zdataset1_reloaded.shape('first') == zdataset1.shape('first')
        assert zdataset1_reloaded.shape('second-') == zdataset2.shape('second')
        assert zdataset1_reloaded.dtype('first') == zdataset1.dtype('first')
        assert zdataset1_reloaded.dtype('second-') == zdataset2.dtype('second')

        a = zdataset1_reloaded.get_array('second-', wrap_with_dask=True).compute()
        b = zdataset2.get_array('second', wrap_with_dask=True).compute()
        assert numpy.all(a == b)

        zdataset1.close()
        zdataset2.close()
        zdataset1_reloaded.close()



from napari import Viewer, gui_qt
from numpy import linspace, dot

from dexp.volumerender.transform_matrices import mat4_translate, mat4_scale, mat4_perspective, mat4_rotation
from dexp.volumerender.volumerender import VolumeRenderer

rend = VolumeRenderer((400, 400))

Nx, Ny, Nz = 200, 150, 50
d = linspace(0, 10000, Nx * Ny * Nz).reshape([Nz, Ny, Nx])

rend.set_data(d)
rend.set_units([1., 1., .1])
rend.set_projection(mat4_perspective(60, 1., 1, 10))
rend.set_modelView(dot(mat4_translate(0, 0, -1), dot(mat4_rotation(1), mat4_scale(.7, .7, .7))))

rend.render()

print(rend.output)

with gui_qt():
    viewer = Viewer()
    viewer.add_image(rend.output, name='output')
    viewer.add_image(rend.output_alpha, name='output_alpha')
    viewer.add_image(rend.output_depth, name='output_depth')



import click
from arbol.arbol import aprint, asection

from dexp.cli.dexp_main import _default_store, _default_codec, _default_clevel
from dexp.cli.dexp_main import _default_workers_backend
from dexp.cli.parsing import _parse_channels, _get_output_path, _parse_slicing
from dexp.datasets.open_dataset import glob_datasets
from dexp.datasets.operations.stabilize import dataset_stabilize


@click.command()
@click.argument('input_paths', nargs=-1)  # ,  help='input path'
@click.option('--output_path', '-o')  # , help='output path'
@click.option('--channels', '-c', default=None, help='List of channels, all channels when omitted.')
@click.option('--slicing', '-s', default=None, help='Dataset slice (TZYX), e.g. [0:5] (first five stacks) [:,0:100] (cropping in z) ')
@click.option('--store', '-st', default=_default_store, help='Zarr store: ‘dir’, ‘ndir’, or ‘zip’', show_default=True)
@click.option('--codec', '-z', default=_default_codec, help='Compression codec: ‘zstd’, ‘blosclz’, ‘lz4’, ‘lz4hc’, ‘zlib’ or ‘snappy’ ', show_default=True)
@click.option('--clevel', '-l', type=int, default=_default_clevel, help='Compression level', show_default=True)
@click.option('--overwrite', '-w', is_flag=True, help='Forces overwrite of target', show_default=True)
@click.option('--maxrange', '-mr', type=int, default=7, help='Maximal distance, in time points, between pairs of images to registrate.', show_default=True)
@click.option('--minconfidence', '-mc', type=float, default=0.5, help='Minimal confidence for registration parameters, if below that level the registration parameters for previous time points is used.', show_default=True)  #
@click.option('--com/--no-com', type=bool, default=False, help='Enable center of mass fallback when standard registration fails.', show_default=True)
@click.option('--quantile', '-q', type=float, default=0.5, help='Quantile to cut-off background in center-of-mass calculation.', show_default=True)
@click.option('--tolerance', '-t', type=float, default=1e-7, help='Tolerance for linear solver.', show_default=True)
@click.option('--ordererror', '-oe', type=float, default=2.0, help='Order for linear solver error term.', show_default=True)
@click.option('--orderreg', '-or', type=float, default=1.0, help='Order for linear solver regularisation term.', show_default=True)
@click.option('--alphareg', '-or', type=float, default=1e-4, help='Multiplicative coefficient for regularisation term.', show_default=True)
@click.option('--pcsigma', '-rs', type=float, default=2, help='Sigma for Gaussian smoothing of phase correlogram, zero to disable.', show_default=True)
@click.option('--dsigma', '-ds', type=float, default=1.5, help='Sigma for Gaussian smoothing (crude denoising) of input images, zero to disable.', show_default=True)
@click.option('--logcomp', '-lc', type=bool, default=True, help='Applies the function log1p to the images to compress high-intensities (usefull when very (too) bright structures are present in the images, such as beads.', show_default=True)
@click.option('--edgefilter', '-ef', type=bool, default=False, help='Applies sobel edge filter to input images.', show_default=True)
@click.option('--workers', '-k', type=int, default=-4, help='Number of worker threads to spawn. Negative numbers n correspond to: number_of _cores / |n|. Be careful, starting two many workers is know to cause trouble (unfortunately unclear why!).',
              show_default=True)
@click.option('--workersbackend', '-wkb', type=str, default=_default_workers_backend, help='What backend to spawn workers with, can be ‘loky’ (multi-process) or ‘threading’ (multi-thread) ', show_default=True)  #
@click.option('--device', '-d', type=int, default=0, help='Sets the CUDA devices id, e.g. 0,1,2', show_default=True)  #
@click.option('--check', '-ck', default=True, help='Checking integrity of written file.', show_default=True)  #
def stabilize(input_paths,
              output_path,
              channels,
              slicing,
              store,
              codec,
              clevel,
              overwrite,
              maxrange,
              minconfidence,
              com,
              quantile,
              tolerance,
              ordererror,
              orderreg,
              alphareg,
              pcsigma,
              dsigma,
              logcomp,
              edgefilter,
              workers,
              workersbackend,
              device,
              check):
    """ Stabilises dataset against translations across time.
    """

    input_dataset, input_paths = glob_datasets(input_paths)
    output_path = _get_output_path(input_paths[0], output_path, "_stabilized")

    slicing = _parse_slicing(slicing)
    channels = _parse_channels(input_dataset, channels)

    with asection(f"Stabilizing dataset(s): {input_paths}, saving it at: {output_path}, for channels: {channels}, slicing: {slicing} "):
        dataset_stabilize(input_dataset,
                          output_path,
                          channels=channels,
                          slicing=slicing,
                          zarr_store=store,
                          compression_codec=codec,
                          compression_level=clevel,
                          overwrite=overwrite,
                          max_range=maxrange,
                          min_confidence=minconfidence,
                          enable_com=com,
                          quantile=quantile,
                          tolerance=tolerance,
                          order_error=ordererror,
                          order_reg=orderreg,
                          alpha_reg=alphareg,
                          phase_correlogram_sigma=pcsigma,
                          denoise_input_sigma=dsigma,
                          log_compression=logcomp,
                          edge_filter=edgefilter,
                          workers=workers,
                          workers_backend=workersbackend,
                          device=device,
                          check=check,
                          debug_output='stabilization'
                          )

        input_dataset.close()
        aprint("Done!")



from typing import Tuple, Optional

import numpy
from arbol import asection
from skimage.data import binary_blobs
from skimage.util import random_noise

from dexp.processing.backends.backend import Backend


def generate_fusion_test_data(length_xy: Optional[int] = 320,
                              length_z_factor: Optional[int] = 4,
                              add_noise: Optional[bool] = True,
                              shift: Optional[Tuple[int, ...]] = None,
                              volume_fraction: Optional[float] = 0.8,
                              amount_low: Optional[float] = 1,
                              zero_level: Optional[float] = 95,
                              odd_dimension: Optional[float] = True,
                              z_overlap: Optional[float] = None,
                              dtype=numpy.float32):
    """

    Parameters
    ----------
    length_xy
    length_z_factor
    add_noise
    shift
    volume_fraction
    amount_low
    zero_level
    odd_dimension
    z_overlap
    dtype

    Returns
    -------

    """
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    with asection("generate blob images"):
        image_gt = binary_blobs(length=length_xy, n_dim=3, blob_size_fraction=0.07, volume_fraction=0.1).astype(dtype)
        blend_a = binary_blobs(length=length_xy, n_dim=3, blob_size_fraction=0.2, volume_fraction=volume_fraction).astype(dtype)
        blend_b = binary_blobs(length=length_xy, n_dim=3, blob_size_fraction=0.2, volume_fraction=volume_fraction).astype(dtype)

    with asection("convert blob images to backend"):
        image_gt = Backend.to_backend(image_gt)
        blend_a = Backend.to_backend(blend_a)
        blend_b = Backend.to_backend(blend_b)

    with asection("prepare high/low image pair"):
        image_gt = sp.ndimage.gaussian_filter(image_gt, sigma=1)
        image_gt = image_gt / xp.max(image_gt)
        image_highq = image_gt.copy()
        image_lowq = image_gt.copy()
        image_lowq = amount_low * sp.ndimage.gaussian_filter(image_lowq, sigma=7)

    with asection("prepare blend maps"):
        blend_a = sp.ndimage.gaussian_filter(blend_a, sigma=2)
        blend_a = blend_a / numpy.max(blend_a)
        blend_b = sp.ndimage.gaussian_filter(blend_b, sigma=2)
        blend_b = blend_b / numpy.max(blend_b)

    if z_overlap is not None:
        with asection("add z bias to blend maps"):
            length = blend_a.shape[0]
            bias_vector = xp.linspace(0, 1, num=length)
            new_shape = tuple(s if i == 0 else 1 for i, s in enumerate(blend_a.shape))
            bias_vector = xp.reshape(bias_vector, newshape=new_shape)
            blend_a *= (1 - bias_vector) ** z_overlap
            blend_b *= (bias_vector) ** z_overlap

    with asection("generate two views via blending"):
        image1 = image_highq * blend_a + image_lowq * blend_b
        image2 = image_highq * blend_b + image_lowq * blend_a

    if length_z_factor != 1:
        with asection("downscale along z"):
            image_gt = sp.ndimage.zoom(image_gt, zoom=(1 / length_z_factor, 1, 1), order=0)
            image_lowq = sp.ndimage.zoom(image_lowq, zoom=(1 / length_z_factor, 1, 1), order=0)
            blend_a = sp.ndimage.zoom(blend_a, zoom=(1 / length_z_factor, 1, 1), order=0)
            blend_b = sp.ndimage.zoom(blend_b, zoom=(1 / length_z_factor, 1, 1), order=0)
            image1 = sp.ndimage.zoom(image1, zoom=(1 / length_z_factor, 1, 1), order=0)
            image2 = sp.ndimage.zoom(image2, zoom=(1 / length_z_factor, 1, 1), order=0)

    if odd_dimension:
        image_gt = xp.pad(image_gt, pad_width=((0, 1), (0, 0), (1, 0)), mode='edge')
        image_lowq = xp.pad(image_lowq, pad_width=((0, 1), (0, 0), (1, 0)), mode='edge')
        blend_a = xp.pad(blend_a, pad_width=((0, 1), (0, 0), (1, 0)), mode='edge')
        blend_b = xp.pad(blend_b, pad_width=((0, 1), (0, 0), (1, 0)), mode='edge')
        image1 = xp.pad(image1, pad_width=((0, 1), (0, 0), (1, 0)), mode='edge')
        image2 = xp.pad(image2, pad_width=((0, 1), (0, 0), (1, 0)), mode='edge')

    if add_noise:
        with asection("add noise"):
            image1 = Backend.to_numpy(image1)
            image2 = Backend.to_numpy(image2)
            image1 = random_noise(image1, mode='speckle', var=0.5)
            image2 = random_noise(image2, mode='speckle', var=0.5)
            image1 = Backend.to_backend(image1)
            image2 = Backend.to_backend(image2)

    if shift is not None:
        with asection("Shift second view relative to first"):
            image2 = sp.ndimage.shift(image2, shift=shift, order=1)

    with asection("scale image intensities"):
        image1 = zero_level + 300 * image1
        image2 = zero_level + 300 * image2
        image_gt = zero_level + 300 * image_gt
        image_lowq = zero_level + 300 * image_lowq

    return image_gt.astype(dtype, copy=False), \
           image_lowq.astype(dtype, copy=False), \
           blend_a.astype(dtype, copy=False), \
           blend_b.astype(dtype, copy=False), \
           image1.astype(dtype, copy=False), \
           image2.astype(dtype, copy=False)



from arbol import aprint

from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend


def BestBackend(*args, **kwargs):
    try:
        import cupy
        with cupy.cuda.Device(0):
            array = cupy.array([1, 2, 3])
            assert cupy.median(array) == 2
        return CupyBackend(*args, **kwargs)

    except Exception:
        aprint("Cupy module not found or not functional! ignored!")
        return NumpyBackend(*args, **kwargs)



from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.filters.sobel_filter import sobel_filter
from dexp.processing.synthetic_datasets.multiview_data import generate_fusion_test_data


def demo_sobel_numpy():
    with NumpyBackend():
        _demo_sobel()


def demo_sobel_cupy():
    try:
        with CupyBackend():
            _demo_sobel()
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def _demo_sobel():
    image_gt, image_lowq, blend_a, blend_b, image1, image2 = generate_fusion_test_data(add_noise=False)

    tenengrad_image1 = sobel_filter(image1)
    tenengrad_image2 = sobel_filter(image2)

    assert tenengrad_image1.shape == image1.shape
    assert tenengrad_image2.shape == image2.shape
    assert tenengrad_image1.dtype == image2.dtype
    assert tenengrad_image2.dtype == image2.dtype

    from napari import Viewer, gui_qt
    with gui_qt():
        def _c(array):
            return Backend.to_numpy(array)

        viewer = Viewer()
        viewer.add_image(_c(image_gt), name='image_gt')
        viewer.add_image(_c(image_lowq), name='image_lowq')
        viewer.add_image(_c(blend_a), name='blend_a')
        viewer.add_image(_c(blend_b), name='blend_b')
        viewer.add_image(_c(image1), name='image1')
        viewer.add_image(_c(image2), name='image2')
        viewer.add_image(_c(tenengrad_image1), name='tenengrad_image1')
        viewer.add_image(_c(tenengrad_image2), name='tenengrad_image2')


if __name__ == "__main__":
    demo_sobel_cupy()
    demo_sobel_numpy()



import os
from os import listdir
from os.path import exists, isfile, join, isdir
from typing import Tuple, Union

import imageio
from arbol.arbol import asection, aprint
from joblib import Parallel, delayed

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.best_backend import BestBackend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.color.crop_resize_pad import crop_resize_pad_color_image


def crop_resize_pad_image_sequence(input_path: str,
                                   output_path: str = None,
                                   crop: Union[int, Tuple[int, ...], Tuple[Tuple[int, int], ...]] = None,
                                   resize: Tuple[int, ...] = None,
                                   resize_order: int = 3,
                                   resize_mode: str = 'constant',
                                   pad_width: Tuple[Tuple[int, int], ...] = None,
                                   pad_mode: str = 'constant',
                                   pad_color: Tuple[float, float, float, float] = (0, 0, 0, 0),
                                   rgba_value_max: float = 255,
                                   overwrite: bool = False,
                                   workers: int = -1,
                                   workersbackend: str = 'threading',
                                   device: int = 0):
    """
    Crops, resizes and then pad a sequence of RGB(A) images.

    Parameters
    ----------
    input_path : Path to folder containing images in some (lexicographic) order.
    output_path : Path to save the blended images.
    crop: Crop image by removing a given number of pixels/voxels per axis. For example: ((10,20),(10,20)) crops 10 pixels on the left for axis 0, 20 pixels from the right of axis 0, and the same for axis 2.
    resize: After cropping, the image is resized to the given shape. If any entry in the tuple is -1 then that position in the shape is automatically determined based on the existing shape to preserve aspect ratio.
    resize_order: The order of the spline interpolation, default is 3. The order has to be in the range 0-5.
    resize_mode: optional The mode parameter determines how the input array is extended beyond its boundaries. Can be: ‘reflect’, ‘constant’, ‘nearest’, ‘mirror’, ‘wrap’.
    pad_width: After cropping and resizing, padding is performed. The provided tuple is interpreted similarly to cropping.
    pad_mode: Padding mode, see numpy.pad for the available modes.
    pad_color: Padding color as tuple of normalised floats:  (R,G,B,A). Default is transparent black.
    rgba_value_max: max value for rgba values.
    overwrite : If True the output files are overwritten
    workers : Number of worker threads to spawn, if -1 then num workers = num devices', show_default=True)
    workersbackend : What backend to spawn workers with, can be ‘loky’ (multi-process) or ‘threading’ (multi-thread)
    device: Device on  which to run the overlay computation (if a non-CPU device is available).
    """

    # accomodate cupy limits:
    if type(Backend.current()) == CupyBackend:
        resize_order = min(resize_order, 1)

    # ensure folder exists:
    os.makedirs(output_path, exist_ok=True)

    # collect image files:
    if isdir(input_path):
        # path is folder:
        png_file_paths = [join(input_path, f) for f in listdir(input_path) if isfile(join(input_path, f)) and f.endswith('.png')]
        png_file_paths.sort()
    else:
        raise ValueError("Input path must be folder containing at least one image")

    if workers == -1:
        workers = min(8, os.cpu_count() // 2)

    # Number of timepoints:
    nb_timepoints = len(png_file_paths)

    def _process(tp: int):
        with asection(f'Processing time point: {tp}'):
            with BestBackend(device, exclusive=True, enable_unified_memory=True):

                # Output file:
                filename = f"frame_{tp:05}.png"
                filepath = join(output_path, filename)

                # Write file:
                if overwrite or not exists(filepath):

                    # Get image path:
                    image_path = png_file_paths[tp]

                    # get image:
                    image = imageio.imread(image_path)

                    resized_image = crop_resize_pad_color_image(image=image,
                                                                crop=crop,
                                                                resize=resize,
                                                                resize_order=resize_order,
                                                                resize_mode=resize_mode,
                                                                pad_width=pad_width,
                                                                pad_mode=pad_mode,
                                                                pad_color=pad_color,
                                                                rgba_value_max=rgba_value_max)

                    with asection(f"Writing file: {filename} in folder: {output_path}"):
                        imageio.imwrite(filepath,
                                        Backend.to_numpy(resized_image),
                                        compress_level=1)
                else:
                    aprint(f"File: {filepath} already exists! use -w option to force overwrite...")

    with asection(f"Cropping ({crop}), resizing ({resize}), padding ({pad_width}), images at: {input_path}, and saving to {output_path}, for a total of {nb_timepoints} time points"):
        Parallel(n_jobs=workers, backend=workersbackend)(delayed(_process)(tp) for tp in range(nb_timepoints))
        aprint(f"Done!")



import gc
from typing import Sequence

import numpy
from arbol import asection, section, aprint

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.equalise.equalise_intensity import equalise_intensity
from dexp.processing.filters.butterworth_filter import butterworth_filter
from dexp.processing.fusion.dct_fusion import fuse_dct_nd
from dexp.processing.fusion.dft_fusion import fuse_dft_nd
from dexp.processing.fusion.tg_fusion import fuse_tg_nd
from dexp.processing.registration.model.pairwise_registration_model import PairwiseRegistrationModel
from dexp.processing.registration.translation_nd import register_translation_nd
from dexp.processing.registration.translation_nd_proj import register_translation_proj_nd
from dexp.processing.restoration.clean_dark_regions import clean_dark_regions
from dexp.processing.restoration.dehazing import dehaze


@section("SimView 2C2L fusion")
def simview_fuse_2C2L(C0L0, C0L1, C1L0, C1L1,
                      equalise: bool = True,
                      equalisation_ratios: Sequence[float] = (None, None, None),
                      zero_level: float = 120,
                      clip_too_high: int = 1024,
                      fusion='tg',
                      fusion_bias_exponent: int = 2,
                      fusion_bias_strength_i: float = 0.5,
                      fusion_bias_strength_d: float = 0.02,
                      registration_mode: str = 'projection',
                      registration_edge_filter: bool = False,
                      registration_crop_factor_along_z: float = 0.3,
                      registration_force_model: bool = False,
                      registration_model: PairwiseRegistrationModel = None,
                      registration_min_confidence: float = 0.5,
                      registration_max_change: int = 16,
                      dehaze_before_fusion: bool = True,
                      dehaze_size: int = 32,
                      dehaze_correct_max_level: bool = True,
                      dark_denoise_threshold: int = 0,
                      dark_denoise_size: int = 9,
                      butterworth_filter_cutoff: float = 1,
                      flip_camera1: bool = True,
                      huge_dataset_mode: bool = False,
                      internal_dtype=numpy.float16):
    """

    Parameters
    ----------
    C0L0 : Image for Camera 0 lightsheet 0
    C0L1 : Image for Camera 0 lightsheet 1
    C1L0 : Image for Camera 1 lightsheet 0
    C1L1 : Image for Camera 1 lightsheet 1

    equalise : Equalise intensity of views before fusion, or not.

    equalisation_ratios: If provided, these ratios are used instead of calculating equalisation values based on the images.
    When fusingh the four images: C0L0, C0L1, C1L0, C1L1, there are three equalisation ratios: (r0, r1, rc),
    following the following definitions: C0L0 ~= r0 * C0L1, C1L0 ~= r1 * C1L1, and  C0lx ~= rc * C1Lx.

    zero_level : Zero level: that's the minimal detector pixel value floor to substract,
    typically for sCMOS cameras the floor is at around 100 (this is to avoid negative values
    due to electronic noise!). Substracting a bit more than that is a good idea to clear out noise
    in the background --  hence the default of 120.

    clip_too_high : clips very high intensities, to avoid loss of precision when converting an internal format such as float16

    fusion : Fusion mode, can be 'tg', 'dct', 'dft'

    fusion_bias_exponent : Exponent for fusion bias

    fusion_bias_strength_i, fusion_bias_strength_d : Strength of fusion bias for fusing views of different _i_llumination, and of different _d_etections. Set to zero to deactivate

    registration_mode : Registration mode, can be: 'projection' or 'full'.
    Projection mode is faster but might have occasionally  issues for certain samples. Full mode is slower and is only recomended as a last resort.

    registration_edge_filter : apply edge filter to help registration

    registration_crop_factor_along_z : How much to center-crop along z to estimate registration parameters between the two fused camera views. A value of 0.3 means cropping by 30% on both ends.

    registration_force_model : Forces the use of the provided model (see below)

    registration_model : Suggested registration model to use the two camera views (C0Lx and C1Lx) -- typically from a previous time-point or another wavelength.
    Used only if its confidence score is higher than that of the computed registration parameters.

    registration_min_confidence : Minimal confidence for registration parameters, if below that level the registration parameters for previous time points is used.

    registration_max_displacement : Maximal change in registration parameters, if above that level the registration parameters for previous time points is used.

    dehaze_before_fusion : Whether to dehaze the views before fusion or to dehaze the fully fused and registered final image.

    dehaze_size : After all fusion and registration, the final image is dehazed to remove
    large-scale background light caused by scattered illumination and out-of-focus light.
    This parameter controls the scale of the low-pass filter used.

    dehaze_correct_max_level : Should the dehazing correct the reduced local max intensity induced by removing the background?

    dark_denoise_threshold : After all fusion and registration, the final image is processed
    to remove any remaining noise in the dark background region (= hurts compression!).

    dark_denoise_size : Controls the scale over which pixels must be below the threshold to
    be categorised as 'dark background'.

    butterworth_filter_cutoff : At the very end, butterworth_filtering may be applied to
    smooth out spurious high frequencies in the final image. A value of 1 means no filtering,
    a value of e.g. 0.5 means cutting off the top 50% higher frequencies, and keeping all
    frequencies below without any change (that's the point of Butterworth filtering).
    WARNING: Butterworth filtering is currently very slow...

    huge_dataset_mode: optimises memory allocation at the detriment of processing speed to tackle really huge datasets.

    internal_dtype : internal dtype


    Returns
    -------
    Fully registered, fused, dehazed 3D image

    """
    xp = Backend.get_xp_module()

    if C0L0.dtype != C0L1.dtype or C0L0.dtype != C1L0.dtype or C0L0.dtype != C1L1.dtype:
        raise ValueError("The four views must have same dtype!")

    if C0L0.shape != C0L1.shape or C0L0.shape != C1L0.shape or C0L0.shape != C1L1.shape:
        raise ValueError("The four views must have same shapes!")

    if type(Backend.current()) is NumpyBackend:
        internal_dtype = numpy.float32

    original_dtype = C0L0.dtype

    with asection(f"Moving C0L0 and C0L1 to backend storage and converting to {internal_dtype}..."):
        C0L0 = Backend.to_backend(C0L0, dtype=internal_dtype, force_copy=False)
        C0L1 = Backend.to_backend(C0L1, dtype=internal_dtype, force_copy=False)
        Backend.current().clear_memory_pool()

    if clip_too_high > 0:
        with asection(f"Clipping intensities above {clip_too_high} for C0L0 & C0L1"):
            C0L0 = xp.clip(C0L0, a_min=0, a_max=clip_too_high, out=C0L0)
            C0L1 = xp.clip(C0L1, a_min=0, a_max=clip_too_high, out=C0L1)
            Backend.current().clear_memory_pool()

    if equalise:
        with asection(f"Equalise intensity of C0L0 relative to C0L1 ..."):
            C0L0, C0L1, ratio0 = equalise_intensity(C0L0, C0L1,
                                                    zero_level=zero_level,
                                                    correction_ratio=equalisation_ratios[0],
                                                    copy=False)
            aprint(f"Equalisation ratio: {ratio0}")
            Backend.current().clear_memory_pool()

    if dehaze_size > 0 and dehaze_before_fusion:
        with asection(f"Dehaze C0L0 and C0L1 ..."):
            C0L0 = dehaze(C0L0.copy(),
                          size=dehaze_size,
                          minimal_zero_level=0,
                          correct_max_level=True)
            C0L1 = dehaze(C0L1.copy(),
                          size=dehaze_size,
                          minimal_zero_level=0,
                          correct_max_level=True)

            # from napari import gui_qt, Viewer
            # with gui_qt():
            #     def _c(array):
            #         return Backend.to_numpy(array)
            #
            #     viewer = Viewer()
            #     viewer.add_image(_c(C0L0), name='C0L0', contrast_limits=(0, 1000))
            #     viewer.add_image(_c(C0L0_dh), name='C0L0_dh', contrast_limits=(0, 1000))
            #     viewer.add_image(_c(C0L1), name='C0L1', contrast_limits=(0, 1000))
            #     viewer.add_image(_c(C0L1_dh), name='C0L1_dh', contrast_limits=(0, 1000))
            #
            # C0L0 = C0L0_dh
            # C0L1 = C0L1_dh
            Backend.current().clear_memory_pool()

    with asection(f"Fuse illumination views C0L0 and C0L1..."):
        C0lx = fuse_illumination_views(C0L0, C0L1,
                                       mode=fusion,
                                       bias_exponent=fusion_bias_exponent,
                                       bias_strength=fusion_bias_strength_i)

        # from napari import gui_qt, Viewer
        # with gui_qt():
        #     def _c(array):
        #         return Backend.to_numpy(array)
        #
        #     viewer = Viewer()
        #     viewer.add_image(_c(C0L0), name='C0L0', contrast_limits=(0, 1000))
        #     viewer.add_image(_c(C0L1), name='C0L1', contrast_limits=(0, 1000))
        #     viewer.add_image(_c(C0lx), name='C0lx', contrast_limits=(0, 1000))

        del C0L0
        del C0L1
        Backend.current().clear_memory_pool()

    with asection(f"Moving C1L0 and C1L1 to backend storage and converting to {internal_dtype}..."):
        C1L0 = Backend.to_backend(C1L0, dtype=internal_dtype, force_copy=False)
        if flip_camera1:
            C1L0 = xp.flip(C1L0, -1)
        C1L1 = Backend.to_backend(C1L1, dtype=internal_dtype, force_copy=False)
        if flip_camera1:
            C1L1 = xp.flip(C1L1, -1)
        Backend.current().clear_memory_pool()

    if clip_too_high > 0:
        with asection(f"Clipping intensities above {clip_too_high} for C0L0 & C0L1"):
            C1L0 = xp.clip(C1L0, a_min=0, a_max=clip_too_high, out=C1L0)
            C1L1 = xp.clip(C1L1, a_min=0, a_max=clip_too_high, out=C1L1)
            Backend.current().clear_memory_pool()

    if equalise:
        with asection(f"Equalise intensity of C1L0 relative to C1L1 ..."):
            C1L0, C1L1, ratio1 = equalise_intensity(C1L0, C1L1,
                                                    zero_level=zero_level,
                                                    correction_ratio=equalisation_ratios[1],
                                                    copy=False)
            Backend.current().clear_memory_pool()
            aprint(f"Equalisation ratio: {ratio1}")

    if dehaze_size > 0 and dehaze_before_fusion:
        with asection(f"Dehaze C1L0 and C1L1 ..."):
            C1L0 = dehaze(C1L0,
                          size=dehaze_size,
                          minimal_zero_level=0,
                          correct_max_level=dehaze_correct_max_level)
            C1L1 = dehaze(C1L1,
                          size=dehaze_size,
                          minimal_zero_level=0,
                          correct_max_level=dehaze_correct_max_level)
            Backend.current().clear_memory_pool()

    with asection(f"Fuse illumination views C1L0 and C1L1..."):
        C1Lx = fuse_illumination_views(C1L0, C1L1,
                                       mode=fusion,
                                       bias_exponent=fusion_bias_exponent,
                                       bias_strength=fusion_bias_strength_i)

        # from napari import gui_qt, Viewer
        # with gui_qt():
        #     def _c(array):
        #         return Backend.to_numpy(array)
        #
        #     viewer = Viewer()
        #     viewer.add_image(_c(C1L0), name='C1L0', contrast_limits=(0, 1000))
        #     viewer.add_image(_c(C1L1), name='C1L1', contrast_limits=(0, 1000))
        #     viewer.add_image(_c(C1Lx), name='C1Lx', contrast_limits=(0, 1000))

        del C1L0
        del C1L1
        Backend.current().clear_memory_pool()

    if equalise:
        with asection(f"Equalise intensity of C0lx relative to C1Lx ..."):
            C0lx, C1Lx, ratioc = equalise_intensity(C0lx, C1Lx,
                                                    zero_level=0,
                                                    correction_ratio=equalisation_ratios[2],
                                                    copy=False)
            aprint(f"Equalisation ratio: {ratioc}")
            Backend.current().clear_memory_pool()

    with asection(f"Register_stacks C0lx and C1Lx ..."):
        C0lx, C1Lx, registration_model = register_detection_views(C0lx, C1Lx,
                                                                  mode=registration_mode,
                                                                  edge_filter=registration_edge_filter,
                                                                  provided_model=registration_model,
                                                                  force_model=registration_force_model,
                                                                  min_confidence=registration_min_confidence,
                                                                  max_change=registration_max_change,
                                                                  crop_factor_along_z=registration_crop_factor_along_z)
        Backend.current().clear_memory_pool()

    with asection(f"Fuse detection views C0lx and C1Lx..."):
        CxLx = fuse_detection_views(C0lx, C1Lx,
                                    mode=fusion,
                                    bias_exponent=fusion_bias_exponent,
                                    bias_strength=fusion_bias_strength_d)
        del C0lx
        del C1Lx
        Backend.current().clear_memory_pool()

    if dehaze_size > 0 and not dehaze_before_fusion:
        with asection(f"Dehaze CxLx ..."):
            CxLx = dehaze(CxLx,
                          size=dehaze_size,
                          minimal_zero_level=0,
                          correct_max_level=dehaze_correct_max_level)
            Backend.current().clear_memory_pool()

    if dark_denoise_threshold > 0:
        with asection(f"Denoise dark regions of CxLx..."):
            CxLx = clean_dark_regions(CxLx,
                                      size=dark_denoise_size,
                                      threshold=dark_denoise_threshold)
            Backend.current().clear_memory_pool()

    # from napari import gui_qt, Viewer
    # with gui_qt():
    #     def _c(array):
    #         return backend.to_numpy(array)
    #     viewer = Viewer()
    #     viewer.add_image(_c(CxLx), name='CxLx', contrast_limits=(0, 1000))
    #     viewer.add_image(_c(CxLx), name='CxLx_dehazed', contrast_limits=(0, 1000))
    #     #viewer.add_image(_c(CxLx_denoised), name='CxLx_denoised', contrast_limits=(0, 1000))

    if 0 < butterworth_filter_cutoff < 1:
        with asection(f"Filter output using a Butterworth filter"):
            cutoffs = (butterworth_filter_cutoff,) * CxLx.ndim
            CxLx = butterworth_filter(CxLx, shape=(31, 31, 31), cutoffs=cutoffs, cutoffs_in_freq_units=False)
            Backend.current().clear_memory_pool()

    with asection(f"Converting back to original dtype..."):
        if original_dtype is numpy.uint16:
            CxLx = xp.clip(CxLx, 0, None, out=CxLx)
        CxLx = CxLx.astype(dtype=original_dtype, copy=False)
        Backend.current().clear_memory_pool()

    gc.collect()
    Backend.current().clear_memory_pool()

    # equalisation ratios found:
    if None in equalisation_ratios:
        equalisation_ratios = (ratio0, ratio1, ratioc)

    return CxLx, registration_model, equalisation_ratios


def fuse_illumination_views(CxL0, CxL1,
                            mode: str = 'tg',
                            smoothing: int = 12,
                            bias_exponent: int = 2,
                            bias_strength: float = 0.1):
    if mode == 'tg':
        fused = fuse_tg_nd(CxL0, CxL1, downscale=2, tenengrad_smoothing=smoothing, bias_axis=2, bias_exponent=bias_exponent, bias_strength=bias_strength)
    elif mode == 'dct':
        fused = fuse_dct_nd(CxL0, CxL1)
    elif mode == 'dft':
        fused = fuse_dft_nd(CxL0, CxL1)

    return fused


def fuse_detection_views(C0Lx, C1Lx,
                         mode: str = 'tg',
                         smoothing: int = 12,
                         bias_exponent: int = 2,
                         bias_strength: float = 0.1):
    if mode == 'tg':
        fused = fuse_tg_nd(C0Lx, C1Lx, downscale=2, tenengrad_smoothing=smoothing, bias_axis=0, bias_exponent=bias_exponent, bias_strength=bias_strength)
    elif mode == 'dct':
        fused = fuse_dct_nd(C0Lx, C1Lx)
    elif mode == 'dft':
        fused = fuse_dft_nd(C0Lx, C1Lx)
    return fused


def register_detection_views(C0Lx, C1Lx,
                             mode='projection',
                             edge_filter=False,
                             integral=True,
                             provided_model=None,
                             force_model=False,
                             min_confidence=0.5,
                             max_change=16,
                             crop_factor_along_z=0.3):
    C0Lx = Backend.to_backend(C0Lx)
    C1Lx = Backend.to_backend(C1Lx)

    aprint(f"Provided registration model: {provided_model}, overall confidence: {0 if provided_model is None else provided_model.overall_confidence()}")

    if force_model and provided_model is not None:
        model = provided_model
    else:

        depth = C0Lx.shape[0]
        crop = int(depth * crop_factor_along_z)
        C0Lx_c = C0Lx[crop:-crop]
        C1Lx_c = C1Lx[crop:-crop]

        if mode == 'projection':
            new_model = register_translation_proj_nd(C0Lx_c, C1Lx_c,
                                                     edge_filter=edge_filter)
        elif mode == 'full':
            new_model = register_translation_nd(C0Lx_c, C1Lx_c,
                                                edge_filter=edge_filter)

        new_model.integral = integral

        aprint(f"Computed registration model: {new_model}, overall confidence: {new_model.overall_confidence()}")

        if provided_model is None or (new_model.overall_confidence() >= min_confidence or new_model.change_relative_to(provided_model) <= max_change):
            model = new_model
        else:
            model = provided_model

    aprint(f"Applying registration model: {model}, overall confidence: {model.overall_confidence()}")
    C0Lx_reg, C1Lx_reg = model.apply_pair(C0Lx, C1Lx)

    return C0Lx_reg, C1Lx_reg, model



import os
import random
import tempfile
from os.path import join

from arbol import aprint, asection
from dask.array.image import imread

from dexp.datasets.operations.projrender import dataset_projection_rendering
from dexp.datasets.zarr_dataset import ZDataset
from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.synthetic_datasets.nuclei_background_data import generate_nuclei_background_data


def demo_projrender_numpy():
    with NumpyBackend():
        _demo_projrender()


def demo_projrender_cupy():
    try:
        with CupyBackend():
            _demo_projrender(length_xy=320, zoom=4)
            return True
    except ModuleNotFoundError:
        aprint("Cupy module not found! demo ignored")
        return False


def _demo_projrender(length_xy=96,
                     zoom=1,
                     n=64,
                     display=True):
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    # generate nuclei image:
    _, _, image = generate_nuclei_background_data(add_noise=False,
                                                  length_xy=length_xy,
                                                  length_z_factor=1,
                                                  independent_haze=True,
                                                  sphere=True,
                                                  zoom=zoom,
                                                  dtype=xp.float32)

    with asection("prepare simulated timelapse:"):
        # move to backend:
        image = Backend.to_backend(image)

        # generate reference 'ground truth' timelapse
        images = (image.copy() for _ in range(n))

        # modify each image:
        images = (sp.ndimage.shift(image, shift=(random.uniform(-1, 1), random.uniform(-1, 1), random.uniform(-1, 1))) for image in images)
        images = (image + random.uniform(-10, 10) for image in images)

        # turn into array:
        images = xp.stack(images)

    with tempfile.TemporaryDirectory() as tmpdir:
        aprint('created temporary directory', tmpdir)

        with asection("Prepare dataset..."):
            input_path = join(tmpdir, 'dataset.zarr')
            dataset = ZDataset(path=input_path,
                               mode='w',
                               store='dir')

            dataset.add_channel(name='channel',
                                shape=images.shape,
                                chunks=(1, 64, 64, 64),
                                dtype=images.dtype)

            dataset.write_array(channel='channel',
                                array=Backend.to_numpy(images))

        with asection("Project..."):
            # output_folder:
            output_path = join(tmpdir, 'projections')

            # Do actual projection:
            dataset_projection_rendering(input_dataset=dataset,
                                         channels=('channel',),
                                         output_path=output_path,
                                         legendsize=0.05)

        # load images into a dask array:
        filename_pattern = os.path.join(output_path, 'frame_*.png')
        images = imread(filename_pattern)

        if display:
            from napari import Viewer, gui_qt
            with gui_qt():
                def _c(array):
                    return Backend.to_numpy(array)

                viewer = Viewer(ndisplay=3)
                viewer.add_image(_c(images), name='images', rgb=True)
                viewer.grid.enabled = True


if __name__ == "__main__":
    if not demo_projrender_cupy():
        demo_projrender_numpy()



# /usr/local/cuda/bin/ncu -o profile python cusparse.py
import cupy
import cupyx


def _make(xp, sp, dtype):
    data = xp.array([0, 1, 3, 2], dtype)
    indices = xp.array([0, 0, 2, 1], 'i')
    indptr = xp.array([0, 1, 2, 3, 4], 'i')
    # 0, 1, 0, 0
    # 0, 0, 0, 2
    # 0, 0, 3, 0
    return sp.csc_matrix((data, indices, indptr), shape=(3, 4))


x = _make(cupy, cupyx.scipy.sparse, float)
expected = cupy.cusparse.cscsort(x)
cupy.cuda.Device().synchronize()

stream = cupy.cuda.stream.Stream()
with stream:
    y = cupy.cusparse.cscsort(x)
stream.synchronize()
cupy.testing.assert_array_equal(y, expected)

stream = cupy.cuda.stream.Stream()
stream.use()
y = cupy.cusparse.cscsort(x)
stream.synchronize()
cupy.testing.assert_array_equal(y, expected)



import click
from arbol.arbol import aprint

from dexp.video.blend import blend_color_image_sequences
from dexp.cli.dexp_main import _default_workers_backend


@click.command()
@click.argument('input_paths', nargs=-1)
@click.option('--output_path', '-o', type=str, default=None, help='Output folder for blended frames.')
@click.option('--modes', '-m', type=str, default='max',
              help='Either one for all images, or one per image in the form of a sequence. Blending modes are: mean, add, satadd, max, alpha. In the add mode the values are clipped. For alpha-blending, we assume that RGB values are not ‘alpha-premultiplied’.',
              show_default=True)
@click.option('--alphas', '-a', type=str, default='1.0',
              help='List of alpha values for each input image',
              show_default=True)
@click.option('--scales', '-s', type=str, default='1.0',
              help='List of scales ‘s’ for each input image, starting with the second one (first image has always a scale of 1). Scaled images must be of same dimension or smaller than the first images. ',
              show_default=True)
@click.option('--translations', '-t', type=str, default='0,0', help='List of translations ‘x0,y0|x1_y1|...’ for each input image, starting with the second one (first image is always at 0,0). If set to 0,0 no translation is performed for any image. ',
              show_default=True)
@click.option('--borderwidth', '-bw', type=int, default=0, help='Border width added to insets. ', show_default=True)
@click.option('--bordercolor', '-bc', type=str, default='1,1,1,1', help='Border color in RGBA format. For example: 1,1,1,1 is white.', show_default=True)
@click.option('--borderover', '-bo', is_flag=True, help='Border is overlayed on top of inset images, without increasing their size. ', show_default=True)
@click.option('--overwrite', '-w', is_flag=True, help='Force overwrite of output images.', show_default=True)
@click.option('--workers', '-k', type=int, default=-1, help='Number of worker threads to spawn, set to -1 for maximum number of workers', show_default=True)  #
@click.option('--workersbackend', '-wkb', type=str, default=_default_workers_backend, help='What backend to spawn workers with, can be ‘loky’ (multi-process) or ‘threading’ (multi-thread) ', show_default=True)  #
@click.option('--device', '-d', type=int, default=0, help='Sets the CUDA devices id, e.g. 0,1,2', show_default=True)  #
def blend(input_paths,
          output_path,
          modes,
          alphas,
          scales,
          translations,
          borderwidth,
          bordercolor,
          borderover,
          overwrite,
          workers,
          workersbackend,
          device):
    """ Blends videos or still images together, using different alphas, blend modes, and possibly with rescaling, translation, and added borders.
    """

    number_of_inputs = len(input_paths)

    if output_path is None:
        output_path = input_paths[0] + '_blend'
    elif output_path.startswith('_'):
        output_path = input_paths[0] + output_path

    if ',' in modes:
        modes = tuple(mode.strip() for mode in modes.split(','))
    else:
        modes = (modes,) * number_of_inputs

    if ',' in alphas:
        alphas = tuple(float(alpha) for alpha in alphas.split(','))
    else:
        alphas = (float(alphas),) * number_of_inputs

    if ',' in scales:
        scales = tuple(float(scale) for scale in scales.split(','))
    else:
        scales = (float(scales),) * number_of_inputs

    if '|' in translations:
        translations = tuple(tuple(float(v) for v in xy.split(',')) for xy in translations.split('|'))
    elif ',' in translations:
        translations = (tuple(float(v) for v in translations.split(',')),) * number_of_inputs
    else:
        translations = ((0, 0),) * number_of_inputs

    bordercolor = tuple(float(v) for v in bordercolor.split(','))

    if len(input_paths) <= 1:
        aprint("Blending requires at least two paths!")
        return

    blend_color_image_sequences(input_paths=input_paths,
                                output_path=output_path,
                                modes=modes,
                                alphas=alphas,
                                scales=scales,
                                translations=translations,
                                border_width=borderwidth,
                                border_color=bordercolor,
                                border_over_image=borderover,
                                overwrite=overwrite,
                                workers=workers,
                                workersbackend=workersbackend,
                                device=device)



from arbol import asection
from skimage.data import astronaut, logo

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.color.insert import insert_color_image


def demo_insert_numpy():
    with NumpyBackend():
        demo_insert()


def demo_insert_cupy():
    try:
        with CupyBackend():
            demo_insert()
        return True
    except (ModuleNotFoundError, NotImplementedError):
        print("Cupy module not found! ignored!")
        return False


def demo_insert(display=True):
    xp = Backend.get_xp_module()

    image_u = Backend.to_backend(astronaut())
    image_v = Backend.to_backend(logo()[::4, ::4, 0:3])

    with asection("insert_bl"):
        insert_bl = insert_color_image(image=image_u,
                                       inset_image=image_v,
                                       translation='bottom_left',
                                       mode='max',
                                       alpha=1)

    with asection("insert_tr_s2"):
        insert_tr_s2 = insert_color_image(image=image_u,
                                          inset_image=image_v,
                                          translation='top_right',
                                          scale=2,
                                          mode='max',
                                          border_color=(1, 0, 0, 1),
                                          border_width=3,
                                          alpha=1)

    with asection("insert_tr_s1h_add"):
        insert_tr_s1h_add = insert_color_image(image=image_u,
                                               inset_image=image_v,
                                               translation='top_right',
                                               scale=1.5,
                                               mode='add',
                                               border_color=(0, 0, 1, 1),
                                               border_width=10,
                                               alpha=1)

    if display:
        from napari import Viewer, gui_qt
        with gui_qt():
            def _c(array):
                return Backend.to_numpy(array)

            viewer = Viewer()
            viewer.add_image(_c(image_u), name='image_u')
            viewer.add_image(_c(image_v), name='image_v')
            viewer.add_image(_c(insert_bl), name='insert_bl')
            viewer.add_image(_c(insert_tr_s2), name='insert_tr_s2')
            viewer.add_image(_c(insert_tr_s1h_add), name='insert_tr_s1h_add')

            viewer.grid.enabled = True


if __name__ == "__main__":
    if not demo_insert_cupy():
        demo_insert_numpy()



import time

from contextlib import contextmanager
from time import time


@contextmanager
def timeit(description: str = '') -> None:
    start = time()
    yield
    elapsed_time = time() - start

    print(f"Elapsed time for '{description}': {elapsed_time} seconds")
    return elapsed_time



import json
from typing import Sequence

from arbol import asection

from dexp.processing.backends.backend import Backend
from dexp.processing.registration.model.pairwise_registration_model import PairwiseRegistrationModel
from dexp.processing.registration.model.sequence_registration_model import SequenceRegistrationModel
from dexp.processing.registration.model.translation_registration_model import TranslationRegistrationModel
from dexp.processing.registration.model.warp_registration_model import WarpRegistrationModel


def from_json(json_str: str):
    xp = Backend.get_xp_module()
    parsed_model = json.loads(json_str)
    if parsed_model['type'] == 'translation':
        model = TranslationRegistrationModel(shift_vector=xp.asarray(parsed_model['translation']),
                                             confidence=xp.asarray(parsed_model['confidence']))
    elif parsed_model['type'] == 'warp':
        model = WarpRegistrationModel(vector_field=xp.asarray(parsed_model['vector_field']),
                                      confidence=xp.asarray(parsed_model['confidence']))

    elif parsed_model['type'] == 'translation_sequence':
        model_list_json = parsed_model['model_list']
        model_list = list(from_json(model_json) for model_json in model_list_json)
        model = SequenceRegistrationModel(model_list=model_list)

    return model


def model_list_from_file(file_path: str):
    with open(file_path, "r") as models_file:
        lines = models_file.readlines()

        model_list = []
        with asection(f"Loading {len(lines)} models from file: {file_path}"):
            for line in lines:
                line = line.strip()
                model = None if line == 'None' else from_json(line)
                model_list.append(model)

        return model_list


def model_list_to_file(file_path: str, model_list: Sequence[PairwiseRegistrationModel]):
    with open(file_path, "w") as models_file:
        lines = []

        with asection(f"Writing {len(model_list)} models to file: {file_path}"):
            for model in model_list:
                line = 'None' if model is None else model.to_json()
                lines.append(line + '\n')

            models_file.writelines(lines)



from abc import ABC, abstractmethod
from typing import Sequence, Tuple, Any

import numpy


class BaseDataset(ABC):

    def __init__(self, dask_backed=False):
        """ Instanciates a Base Dataset

        """
        self.dask_backed = dask_backed

    def _selected_channels(self, channels: Sequence[str]):
        if channels is None:
            selected_channels = self.channels()
        else:
            selected_channels = [channel for channel in channels if channel in self.channels()]

        # aprint(f"Available channels: {self._channels}")
        # aprint(f"Requested channels: {channels if channels else '--All--'} ")
        # aprint(f"Selected channels:  {selected_channels}")

        return selected_channels

    def _get_largest_dtype_value(self, dtype):
        if numpy.issubdtype(dtype, numpy.integer):
            return numpy.iinfo(dtype).max
        elif numpy.issubdtype(dtype, numpy.float):
            return numpy.finfo(dtype).max

    @abstractmethod
    def close(self):
        pass

    @abstractmethod
    def channels(self) -> Sequence[str]:
        pass

    def nb_timepoints(self, channel: str) -> int:
        return self.shape(channel)[0]

    @abstractmethod
    def shape(self, channel: str) -> Sequence[int]:
        pass

    @abstractmethod
    def dtype(self, channel: str):
        pass

    def tree(self) -> str:
        tree_str = f"{type(self).__name__} dataset"
        tree_str += "\n\n"
        tree_str += "Channels: \n"
        for channel in self.channels():
            tree_str += "  └──" + self.info(channel) + "\n"
        return tree_str

    @abstractmethod
    def info(self, channel: str = None) -> str:
        pass

    @abstractmethod
    def get_metadata(self):
        pass

    @abstractmethod
    def add_channel(self, name: str, shape: Tuple[int, ...], dtype, enable_projections: bool = True, **kwargs) -> Any:
        pass

    @abstractmethod
    def get_array(self, channel: str, per_z_slice: bool = False, wrap_with_dask: bool = False):
        pass

    @abstractmethod
    def get_stack(self, channel: str, time_point: int, per_z_slice: bool = False, wrap_with_dask: bool = False):
        pass

    @abstractmethod
    def get_projection_array(self, channel: str, axis: int, wrap_with_dask: bool = False) -> Any:
        pass

    @abstractmethod
    def write_array(self, channel: str, array: numpy.ndarray):
        pass

    @abstractmethod
    def write_stack(self, channel: str, time_point: int, stack: numpy.ndarray):
        pass

    @abstractmethod
    def check_integrity(self, channels: Sequence[str]) -> bool:
        pass



import click
from arbol.arbol import aprint, asection

from dexp.cli.dexp_main import _default_codec, _default_store, _default_clevel
from dexp.cli.dexp_main import _default_workers_backend
from dexp.cli.parsing import _parse_channels, _get_output_path, _parse_slicing, _parse_devices
from dexp.datasets.open_dataset import glob_datasets
from dexp.datasets.operations.deskew import dataset_deskew


@click.command()
@click.argument('input_paths', nargs=-1)  # ,  help='input path'
@click.option('--output_path', '-o')  # , help='output path'
@click.option('--channels', '-c', default=None, help='list of channels for the view in standard order for the microscope type (C0L0, C0L1, C1L0, C1L1,...)')
@click.option('--slicing', '-s', default=None, help='dataset slice (TZYX), e.g. [0:5] (first five stacks) [:,0:100] (cropping in z) ')  #
@click.option('--store', '-st', default=_default_store, help='Zarr store: ‘dir’, ‘ndir’, or ‘zip’', show_default=True)
@click.option('--codec', '-z', default=_default_codec, help='compression codec: ‘zstd’, ‘blosclz’, ‘lz4’, ‘lz4hc’, ‘zlib’ or ‘snappy’ ')
@click.option('--clevel', '-l', type=int, default=_default_clevel, help='Compression level', show_default=True)
@click.option('--overwrite', '-w', is_flag=True, help='to force overwrite of target', show_default=True)  # , help='dataset slice'
@click.option('--zerolevel', '-zl', type=int, default=110, help="‘zero-level’ i.e. the pixel values in the restoration (to be substracted)", show_default=True)  #
@click.option('--cliphigh', '-ch', type=int, default=1024, help='Clips voxel values above the given value, if zero no clipping is done', show_default=True)  #
@click.option('--dehaze_size', '-dhs', type=int, default=65, help='Filter size (scale) for dehazing the final regsitered and fused image to reduce effect of scattered and out-of-focus light. Set to zero to deactivate.',
              show_default=True)  #
@click.option('--dark_denoise_threshold', '-ddt', type=int, default=0, help='Threshold for denoises the dark pixels of the image -- helps increase compression ratio. Set to zero to deactivate.', show_default=True)  #
@click.option('--workers', '-k', type=int, default=-1, help='Number of worker threads to spawn, if -1 then num workers = num devices', show_default=True)  #
@click.option('--workersbackend', '-wkb', type=str, default=_default_workers_backend, help='What backend to spawn workers with, can be ‘loky’ (multi-process) or ‘threading’ (multi-thread) ', show_default=True)  #
@click.option('--devices', '-d', type=str, default='0', help='Sets the CUDA devices id, e.g. 0,1,2 or ‘all’', show_default=True)  #
@click.option('--check', '-ck', default=True, help='Checking integrity of written file.', show_default=True)  #
def deskew(input_paths,
           output_path,
           channels,
           slicing,
           store,
           codec,
           clevel,
           overwrite,
           zerolevel,
           cliphigh,
           dehaze_size,
           dark_denoise_threshold,
           workers,
           workersbackend,
           devices,
           check):
    """ Deskews all or selected channels of a dataset.
    """

    input_dataset, input_paths = glob_datasets(input_paths)
    output_path = _get_output_path(input_paths[0], output_path, "_deconv")

    slicing = _parse_slicing(slicing)
    channels = _parse_channels(input_dataset, channels)
    devices = _parse_devices(devices)

    with asection(f"Deskewing dataset: {input_paths}, saving it at: {output_path}, for channels: {channels}, slicing: {slicing} "):
        aprint(f"Devices used: {devices}, workers: {workers} ")
        dataset_deskew(input_dataset,
                       output_path,
                       channels=channels,
                       slicing=slicing,
                       store=store,
                       compression=codec,
                       compression_level=clevel,
                       overwrite=overwrite,
                       zero_level=zerolevel,
                       clip_too_high=cliphigh,
                       dehaze_size=dehaze_size,
                       dark_denoise_threshold=dark_denoise_threshold,
                       workers=workers,
                       workersbackend=workersbackend,
                       devices=devices,
                       check=check)

        input_dataset.close()
        aprint("Done!")



from pytest import approx

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.registration.demo.demo_translation_2d import _register_translation_2d


def test_register_translation_2d_numpy():
    with  NumpyBackend():
        register_translation_2d()


def test_register_translation_2d_cupy():
    try:
        with CupyBackend():
            register_translation_2d()
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def register_translation_2d():
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    image, shifted, unshifted, model = _register_translation_2d(display=False)

    shifts = Backend.to_numpy(model.shift_vector)
    assert shifts[0] == approx(-13, abs=0.5)
    assert shifts[1] == approx(5, abs=0.5)

    error_shifted = xp.mean(xp.absolute(image - shifted))
    error_unshifted = xp.mean(xp.absolute(image - unshifted))
    print(f"error_shifted = {error_shifted}")
    print(f"error_unshifted = {error_unshifted}")

    assert error_unshifted < error_shifted
    assert error_unshifted < 0.02



from typing import Sequence, Tuple

from arbol import aprint, asection
from zarr.errors import ArrayNotFoundError

from dexp.datasets.base_dataset import BaseDataset


def dataset_view_remote(input_paths: BaseDataset,
                        name: str,
                        aspect: float,
                        channels: Sequence[str],
                        contrast_limits: Tuple[float, float],
                        colormap: str,
                        slicing,
                        windowsize: int):
    # Annoying napari induced warnings:
    import warnings
    warnings.filterwarnings("ignore")

    if channels is None:
        aprint("Channel(s) must be specified!")
        return

    input_path = input_paths[0]

    if ':' not in input_path.replace("http://", ""):
        input_path = f'{input_path}:8000'

    with asection(f"Viewing remote dataset at: {input_path}, channel(s): {channels}"):
        channels = tuple(channel.strip() for channel in channels.split(','))
        channels = list(set(channels))

        from napari import Viewer, gui_qt

        with gui_qt():
            viewer = Viewer(title=f"DEXP | viewing with napari: {name} ", ndisplay=2)
            viewer.window.resize(windowsize + 256, windowsize)

            for channel in channels:
                import dask.array as da
                if '/' in channel:
                    array = da.from_zarr(f"{input_path}/{channel}")
                else:
                    array = da.from_zarr(f"{input_path}/{channel}/{channel}")

                if slicing is not None:
                    array = array[slicing]

                layer = viewer.add_image(array,
                                         name=channel,
                                         visible=True,
                                         contrast_limits=contrast_limits,
                                         blending='additive',
                                         colormap=colormap,
                                         attenuation=0.04,
                                         rendering='attenuated_mip')

                if aspect is not None:
                    if array.ndim == 3:
                        layer.scale = (aspect, 1, 1)
                    elif array.ndim == 4:
                        layer.scale = (1, aspect, 1, 1)
                    aprint(f"Setting aspect ratio to {aspect} (layer.scale={layer.scale})")

                try:
                    for axis in range(array.ndim):
                        if '/' in channel:
                            proj_array = da.from_zarr(f"{input_path}/{channel}_max{axis}")
                        else:
                            proj_array = da.from_zarr(f"{input_path}/{channel}/{channel}_max{axis}")

                        proj_layer = viewer.add_image(proj_array,
                                                      name=f'{channel}_max{axis}',
                                                      visible=True,
                                                      contrast_limits=contrast_limits,
                                                      blending='additive',
                                                      colormap=colormap)

                        if aspect is not None:
                            if axis == 0:
                                proj_layer.scale = (1, 1)
                            elif axis == 1:
                                proj_layer.scale = (aspect, 1)
                            elif axis == 2:
                                proj_layer.scale = (aspect, 1)

                            aprint(f"Setting aspect ratio for projection (layer.scale={proj_layer.scale})")

                except (KeyError, ArrayNotFoundError):
                    aprint("Warning: could not find projections in dataset!")



import numpy

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.synthetic_datasets.multiview_data import generate_fusion_test_data
from dexp.processing.utils.element_wise_affine import element_wise_affine


def test_element_wise_affine_numpy():
    with NumpyBackend():
        _test_element_wise_affine()


def test_element_wise_affine_cupy():
    try:
        with CupyBackend():
            _test_element_wise_affine()
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def _test_element_wise_affine(length_xy=128):
    xp = Backend.get_xp_module()

    _, _, _, _, image, _ = generate_fusion_test_data(add_noise=False,
                                                     length_xy=length_xy,
                                                     length_z_factor=4)

    transformed = element_wise_affine(image, 2, 0.3)

    transformed = Backend.to_numpy(transformed)
    image = Backend.to_numpy(image)
    error = numpy.median(numpy.abs(image * 2 + 0.3 - transformed))
    print(f"error={error}")
    assert error < 22

    # from napari import Viewer, gui_qt
    # with gui_qt():
    #     viewer = Viewer()
    #     viewer.add_image(image_gt, name='image_gt')
    #     viewer.add_image(image1, name='image1')
    #     viewer.add_image(image2, name='image2')
    #     viewer.add_image(blend_a, name='blend_a')
    #     viewer.add_image(blended, name='blended')



from dexp.processing.backends.backend import Backend
from dexp.processing.backends.numpy_backend import NumpyBackend


def test_backend_context():
    Backend.reset()
    assert type(Backend.current(raise_error_if_none=False)) == NumpyBackend

    with NumpyBackend() as backend_1:
        backend_c = Backend.current()
        assert backend_c == backend_1

        with NumpyBackend() as backend_2:
            backend_c = Backend.current()
            assert backend_c == backend_2
            assert backend_1 != backend_2

        backend_c = Backend.current()
        assert backend_c == backend_1



from arbol import asection, aprint
from napari import gui_qt, Viewer

from dexp.datasets.zarr_dataset import ZDataset
from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.multiview_lightsheet.fusion.simview import simview_fuse_2C2L

filepath = '/mnt/raid0/dexp_datasets/photomanip/ch0.zarr'

zdataset = ZDataset(path=filepath, mode='r')

aprint(zdataset.channels())

C0L0 = zdataset.get_stack('C0L0', 4)
C0L1 = zdataset.get_stack('C0L1', 4)
C1L0 = zdataset.get_stack('C1L0', 4)
C1L1 = zdataset.get_stack('C1L1', 4)

with CupyBackend() as backend:
    with asection("simview_fuse_2I2D"):
        CxLx, model = simview_fuse_2C2L(C0L0, C0L1, C1L0, C1L1,
                                        clip_too_high=1024)
    aprint(f"Model = {model}")

    with asection("to_numpy"):
        CxLx = Backend.to_numpy(CxLx)

    with gui_qt():
        def _c(array):
            return Backend.to_numpy(array)


        viewer = Viewer()
        viewer.add_image(_c(C0L0), name='C0L0', contrast_limits=(0, 1000), scale=(4, 1, 1), blending='additive', visible=False)
        viewer.add_image(_c(C0L1), name='C0L1', contrast_limits=(0, 1000), scale=(4, 1, 1), blending='additive', visible=False)
        viewer.add_image(_c(C1L0), name='C1L0', contrast_limits=(0, 1000), scale=(4, 1, 1), blending='additive', visible=False)
        viewer.add_image(_c(C1L1), name='C1L1', contrast_limits=(0, 1000), scale=(4, 1, 1), blending='additive', visible=False)
        viewer.add_image(_c(CxLx), name='CxLx', contrast_limits=(0, 1200), scale=(4, 1, 1), blending='additive', colormap='viridis')
        # viewer.add_image(_c(CxLx_deconvolved), name='CxLx_deconvolved', contrast_limits=(0, 1000), scale=(4, 1, 1), blending='additive', colormap='viridis')



import tempfile
from os.path import join

import napari
from napari import Viewer
from skimage.data import binary_blobs
from skimage.filters import gaussian

from dexp.datasets.zarr_dataset import ZDataset


def demo():
    with tempfile.TemporaryDirectory() as tmpdir:
        print('created temporary directory', tmpdir)

        zdataset = ZDataset(path=join(tmpdir, 'test.zarr'),
                            mode='w',
                            store='dir')

        array1 = zdataset.add_channel(name='first',
                                      shape=(10, 100, 100, 100),
                                      chunks=(1, 50, 50, 50),
                                      dtype='f4',
                                      codec='zstd',
                                      clevel=3)

        array2 = zdataset.add_channel(name='second',
                                      shape=(17, 10, 20, 30),
                                      chunks=(1, 5, 1, 6),
                                      dtype='f4',
                                      codec='zstd',
                                      clevel=3)

        for i in range(0, 10):
            blobs = binary_blobs(length=100, n_dim=3, blob_size_fraction=0.1).astype('f4')
            blobs = gaussian(blobs, sigma=1)
            array1[i] = blobs

        for i in range(0, 17):
            blobs = binary_blobs(length=30, n_dim=3, blob_size_fraction=0.03).astype('f4')
            blobs = gaussian(blobs, sigma=1)
            array2[i] = blobs[0:10, 0:20, 0:30]

        print(array1.info)
        print(array2.info)

        with napari.gui_qt():
            viewer = Viewer()
            viewer.add_image(array1, name='array_first')
            viewer.add_image(array2, name='array_second')

        zdataset.close()
        del zdataset
        del array1
        del array2

        zdataset_read = ZDataset(path=join(tmpdir, 'test.zarr'),
                                 mode='r')

        array1 = zdataset_read.get_array('first')
        array2 = zdataset_read.get_array('second')

        print(array1.info)
        print(array2.info)

        with napari.gui_qt():
            viewer = Viewer()
            viewer.add_image(array1, name='array_first')
            viewer.add_image(array2, name='array_second')


demo()



from typing import Callable

import numpy
from arbol import aprint
from numpy import ndarray
from telegram import Update
from telegram.ext import Updater, CommandHandler, MessageHandler, Filters, CallbackContext, Handler
# Define a few command handlers. These usually take the two arguments update and
# context. Error handlers also receive the raised TelegramError object in error.
from telegram.ext.dispatcher import DEFAULT_GROUP

from dexp.cli.config import get_telegram_bot_token, get_telegram_bot_password
from dexp.processing.backends.backend import Backend


# import logging
# # Enable logging
# from dexp.cli.config import get_telegram_bot_token
#
# logging.basicConfig(
#     format='%(asctime)s - %(name)s - %(levelname)s - %(message)s', level=logging.INFO
# )
#
# logger = logging.getLogger(__name__)


class TelegramBot:

    def __init__(self, token: str = get_telegram_bot_token(), password: str = get_telegram_bot_password()):
        """
        Constructs a Telegram Bot

        Parameters
        ----------
        blocking : if True then this call will block until the Ctrl-C is pressed or the process receives SIGINT, SIGTERM or SIGABRT
        """

        super().__init__()

        self.updater = Updater(token=token, use_context=True)
        self.password: str = password
        self.authorised: bool = False

        def password_command(update: Update, context: CallbackContext) -> None:
            """Send a message when the command /help is issued."""
            if self.password in update.message.text:
                update.message.reply_text('✅ Authentication successful!')
                self.authorised = True
            else:
                update.message.reply_text('❌ Authentication failed!')

        self._add_handler(CommandHandler("password", password_command))

    def _add_handler(self, handler: Handler, group: int = DEFAULT_GROUP):
        self.updater.dispatcher.add_handler(handler=handler, group=group)

    def add_text_handler(self, function: Callable[[str], str]):
        """
        Adds a simple handler that takes a text message and returns a text message in return.

        Parameters
        ----------
        function : function that takes a string and returns a string
        """

        def _fun(update: Update, context: CallbackContext) -> None:
            if self.authorised:
                message = update.message.text
                reply = function(message)
                aprint(f"Telegram Bot: received '{message}', replying: '{reply}'")
                update.message.reply_text(reply)

        self._add_handler(MessageHandler(Filters.text & ~Filters.command, _fun))

    def add_image_handler(self, function: Callable[[str], ndarray]):
        """
        Adds a simple handler that takes a text message as input and returns an image in return.
        The image is passed as a 2D numpy array normalised within [0, 1]

        Parameters
        ----------
        function : function that takes a string and returns an image (2D numpy array)
        """

        def _fun(update: Update, context: CallbackContext) -> None:
            if self.authorised:
                message = update.message.text
                array, cmap_name = function(message)
                array = Backend.to_numpy(array)

                from matplotlib.cm import get_cmap
                cmap = get_cmap(cmap_name)

                from PIL import Image
                image = Image.fromarray(numpy.uint8(cmap(array)[..., 0:3] * 255), mode='RGB')
                from io import BytesIO
                image_bytes = BytesIO()
                image_bytes.name = 'image.png'
                image.save(image_bytes, 'PNG')
                image_bytes.seek(0)

                aprint(f"Telegram Bot: received '{message}', replying with image: '{image}'")
                update.message.reply_photo(image_bytes)

        self._add_handler(MessageHandler(Filters.text & ~Filters.command, _fun))

    def start(self, blocking=False, clean=True):
        """
        Starts the bot

        Parameters
        ----------
        blocking : if True then this call will block until the Ctrl-C is pressed or the process receives SIGINT, SIGTERM or SIGABRT
        """
        self.updater.start_polling(clean=clean)

        if blocking:
            self.idle()

    def idle(self):
        """
        Run the bot until you press Ctrl-C or the process receives SIGINT, SIGTERM or SIGABRT. This should be used most of the time, since
        start_polling() is non-blocking and will stop the bot gracefully.

        """

        self.updater.idle()

    def close(self):
        """
        Close (stops) the bot

        """
        self.updater.stop()



import numpy

from dexp.datasets.zarr_dataset import ZDataset
from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.fusion.tg_fusion import fuse_tg_nd
from dexp.processing.registration.warp_multiscale_nd import register_warp_multiscale_nd
from dexp.processing.restoration.dehazing import dehaze
from dexp.processing.utils.normalise import normalise_functions
from dexp.utils.timeit import timeit

dataset1_path = '/mnt/raid0/pisces_datasets/488beads_TL100_range1000um_step0.31_4um_20ms_dualv_1tp_without_stripeReduc_resampled_intervel_1.zarr'
dataset2_path = '/mnt/raid0/pisces_datasets/data1_beads_TL100_range1200um_step0.31_6um_50ms_dualv_1tp_2_resampled_intervel_1.zarr/'
dataset3_path = '/mnt/raid0/pisces_datasets/data2_fish_TL100_range1300um_step0.31_6um_20ms_dualv_300tp_2_resampled_intervel_1_first10tp.zarr'
dataset4_path = '/mnt/raid0/dexp_datasets/tail/raw.zarr'

zdataset = ZDataset(path=dataset4_path, mode='r')

print(zdataset.channels())

tp = 3

view1 = zdataset.get_stack('v0c0_rot', tp)  # [..., 0:-500]
view2 = zdataset.get_stack('v1c0_rot', tp)  # [..., 500:]

# view1 = view1[:,1765-256:1765+256, 1929-1024:1929+1024]
# view2 = view2[:,1765-256:1765+256, 1929-1024:1929+1024]

print(f"view1 shape={view1.shape}, dtype={view1.dtype}")
print(f"view2 shape={view2.shape}, dtype={view2.dtype}")

with CupyBackend(0, enable_memory_pool=False) as backend:
    print(backend)

    with timeit("dehaze"):
        view1 = dehaze(view1, size=25, in_place=True)
        view2 = dehaze(view2, size=25, in_place=True)

    with timeit("to_backend"):
        view1 = Backend.to_backend(view1, dtype=numpy.float32)
        view2 = Backend.to_backend(view2, dtype=numpy.float32)

    with timeit("normalise"):
        norm_fun1, _ = normalise_functions(view1, low=0, high=1024, quantile=0.0001)
        norm_fun2, _ = normalise_functions(view2, low=0, high=1024, quantile=0.0001)

        view1 = norm_fun1(view1)
        view2 = norm_fun2(view2)

    print(f"1: min={view1.min()}, max={view1.max()}")
    print(f"2: min={view2.min()}, max={view2.max()}")

    # from napari import Viewer, gui_qt
    # with gui_qt():
    #     def _c(array):
    #         return backend.to_numpy(array)
    #
    #     viewer = Viewer()
    #
    #     viewer.add_image(_c(view1), name='view1', colormap='bop blue', blending='additive')
    #     viewer.add_image(_c(view2), name='view2', colormap='bop orange', blending='additive')

    edge_filter = False

    with timeit("register_warp_multiscale_nd"):
        model = register_warp_multiscale_nd(view1, view2,
                                            num_iterations=5,
                                            confidence_threshold=0.3,
                                            edge_filter=edge_filter,
                                            denoise_input_sigma=1)

    with timeit("unwarp"):
        model.integral = True
        view1_reg, view2_reg = model.apply_pair(view1, view2)

    view1 = Backend.to_numpy(view1)
    view2 = Backend.to_numpy(view2)

    with timeit("fuse"):
        fused = fuse_tg_nd(view1_reg, view2_reg)

    from napari import Viewer, gui_qt

    with gui_qt():
        def _c(array):
            return Backend.to_numpy(array)


        viewer = Viewer()
        viewer.add_image(_c(view1), name='view1', colormap='bop blue', blending='additive')
        viewer.add_image(_c(view2), name='view2', colormap='bop orange', blending='additive')
        viewer.add_image(_c(view2_reg), name='view2_reg', colormap='bop orange', blending='additive')
        viewer.add_image(_c(fused), name='fused', colormap='bop purple', blending='additive')
        viewer.add_image(_c(model.vector_field[..., 0]), name='vx', colormap='viridis', blending='additive')
        viewer.add_image(_c(model.vector_field[..., 1]), name='vy', colormap='viridis', blending='additive')
        viewer.add_image(_c(model.vector_field[..., 2]), name='vz', colormap='viridis', blending='additive')



import numpy

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.utils.normalise import normalise_functions


def sobel_filter(image,
                 exponent: int = 2,
                 gamma: float = 1,
                 log_compression: bool = True,
                 normalise_input: bool = True,
                 in_place_normalisation: bool = False,
                 internal_dtype=None):
    """
    Computes the Sobel magnitude filter response for a given image.

    The Sobel operator, sometimes called the Sobel–Feldman operator or Sobel filter,
    is used in image processing and computer vision, particularly within edge detection
    algorithms where it creates an image emphasising edges. It is named after Irwin Sobel
    and Gary Feldman, colleagues at the Stanford Artificial Intelligence Laboratory (SAIL).
    Sobel and Feldman presented the idea of an "Isotropic 3x3 Image Gradient Operator" at
    a talk at SAIL in 1968. Technically, it is a discrete differentiation operator,
    computing an approximation of the gradient of the image intensity function.
    At each point in the image, the result of the Sobel–Feldman operator is either the
    corresponding gradient vector or the norm of this vector. The Sobel–Feldman operator
    is based on convolving the image with a small, separable, and integer-valued filter
    in the horizontal and vertical directions and is therefore relatively inexpensive in
    terms of computations. On the other hand, the gradient approximation that it produces
    is relatively crude, in particular for high-frequency variations in the image.

    (Source: wikipedia, https://en.wikipedia.org/wiki/Sobel_operator)

    Parameters
    ----------
    image : image to apply filter on
    exponent : Exponent to use for the magnitude (norm) of the gradient, 2 for L2, and 1 for L1...
    gamma : After normalisation, applies thte given gamma value.
    log_compression : Before normalisation, applies log compression -- usefull to reduce the impact of high intensity values versus contrast.
    normalise_input : True to normalise input image between 0 and 1 before applying filter
    in_place_normalisation : If True then input image can be modified during normalisation.
    internal_dtype : dtype for internal computation.

    Returns
    -------
    Filtered image

    """
    xp = Backend.get_xp_module(image)
    sp = Backend.get_sp_module(image)
    ndim = image.ndim

    if internal_dtype is None:
        internal_dtype = image.dtype

    if type(Backend.current()) is NumpyBackend:
        internal_dtype = numpy.float32

    original_dtype = image.dtype
    force_copy = normalise_input and not in_place_normalisation
    image = Backend.to_backend(image, dtype=internal_dtype, force_copy=force_copy)

    if log_compression:
        image = xp.log1p(image, out=image if force_copy else None)

    norm_fun, denorm_fun = normalise_functions(image, do_normalise=normalise_input, dtype=internal_dtype)

    image = norm_fun(image)

    if gamma != 1:
        image **= gamma

    sobel_image = xp.zeros_like(image)

    for i in range(ndim):
        sobel_one_axis = xp.absolute(sp.ndimage.sobel(image, axis=i))
        sobel_image += sobel_one_axis ** exponent

    if exponent == 1:
        pass
    elif exponent == 2:
        sobel_image = xp.sqrt(sobel_image, out=sobel_image)

    sobel_image = sobel_image.astype(dtype=original_dtype, copy=False)

    return sobel_image



from typing import Union, Tuple

import numpy

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.numpy_backend import NumpyBackend


def gaussian_kernel_nd(size: Union[int, Tuple[int, ...]] = 5,
                       ndim: int = None,
                       sigma: float = 1.0,
                       dtype=numpy.float16):
    """
    Computes a nD Gaussian kernel
    Parameters
    ----------
    ndim : number of dimensions
    size : size in pixels
    sigma : Gaussian sigma
    dtype : dtype for kernel

    Returns
    -------
    nD Gaussian kernel

    """
    backend = Backend.current()
    xp = backend.get_xp_module()
    sp = backend.get_sp_module()

    if type(backend) is NumpyBackend:
        dtype = numpy.float32

    if type(size) is not tuple and ndim is not None:
        size = (size,) * ndim

    kernel = xp.zeros(shape=size, dtype=dtype)
    slicing = tuple(slice(s // 2, s // 2 + 1, None) for s in size)
    kernel[slicing] = 1
    kernel = sp.ndimage.gaussian_filter(kernel, sigma=sigma)

    kernel /= xp.sum(kernel)

    return kernel



from arbol import aprint

from dexp.datasets.operations.demo.demo_stabilize import _demo_stabilize
from dexp.processing.backends.cupy_backend import CupyBackend


# def test_stabilize_numpy():
#     with NumpyBackend():
#         _demo_stabilize(display=False)


def test_stabilize_cupy():
    try:
        with CupyBackend():
            _demo_stabilize(display=False)

    except ModuleNotFoundError:
        aprint("Cupy module not found! demo ignored")



from typing import Tuple, Union, Optional

import numpy

from dexp.processing.backends.backend import Backend
from dexp.processing.utils.nd_slice import nd_split_slices, remove_margin_slice
from dexp.processing.utils.normalise import normalise_functions


def scatter_gather_i2i(function,
                       image,
                       chunks: Union[int, Tuple[int, ...]],
                       margins: Optional[Union[int, Tuple[int, ...]]] = None,
                       normalise: bool = False,
                       clip: bool = False,
                       to_numpy: bool = True,
                       internal_dtype=None):
    """
    Image-2-image scatter-gather.
    'Scatters' computation of a given unary function by splitting the input array into chunks, computing using a given backend,
    and reassembling the chunks into a single array of same shape as the inpout that is either backed by the same backend than
    that of the input image, or that is backed by numpy -- usefull when the compute backend cannot hold the whole input and output
    images in memory.

    Parameters
    ----------
    function : unary function
    image : input image (can be any backend, numpy )
    chunks : chunk sizes to cut input image into, can be a single integer or a tuple of integers.
    margins : margins to add to each chunk, can be a single integer or a tuple of integers. if None, no margins are added.
    normalise : normalises  the input image.
    clip : clip after normalisation/denormalisation
    to_numpy : should the result be a numpy array? Very usefull when the compute backend cannot hold the whole input and output images in memory.
    internal_dtype : internal dtype for computation

    Returns
    -------
    Result of applying the unary function to the input image, if to_numpy==True then the image is

    """

    if internal_dtype is None:
        internal_dtype = image.dtype

    if type(chunks) == int:
        chunks = (chunks,) * image.ndim

    # If None is passed for a chunk that means that we don't chunk along that axis, we als clip the chunk size:
    chunks = tuple((length if chunk is None else min(length, chunk)) for chunk, length in zip(chunks, image.shape))

    if margins is None:
        margins = (0,) * image.ndim

    if type(margins) == int:
        margins = (margins,) * image.ndim

    if to_numpy:
        result = numpy.empty(shape=image.shape, dtype=image.dtype)
    else:
        result = Backend.get_xp_module(image).empty_like(image)

    # Normalise:
    norm_fun, denorm_fun = normalise_functions(image, do_normalise=normalise, clip=clip)

    # image shape:
    shape = image.shape

    # We compute the slices objects to cut the input and target images into batches:
    chunk_slices = list(nd_split_slices(shape, chunks=chunks, margins=margins))
    chunk_slices_no_margins = list(nd_split_slices(shape, chunks=chunks))

    # Zipping together slices with and without margins:
    slices = zip(chunk_slices, chunk_slices_no_margins)

    # Number of tiles:
    number_of_tiles = len(chunk_slices)

    if number_of_tiles == 1:
        # If there is only one tile, let's not be complicated about it:
        result = denorm_fun(function(norm_fun(image)))
        if to_numpy:
            result = Backend.to_numpy(result, dtype=image.dtype)
        else:
            result = Backend.to_backend(result, dtype=image.dtype)
    else:
        _scatter_gather_loop(denorm_fun, function, image, internal_dtype, norm_fun, result, shape, slices, to_numpy)

    return result


def _scatter_gather_loop(denorm_fun, function, image, internal_dtype, norm_fun, result, shape, slices, to_numpy):
    for chunk_slice, chunk_slice_no_margins in slices:
        image_chunk = image[chunk_slice]
        image_chunk = Backend.to_backend(image_chunk, dtype=internal_dtype)
        image_chunk = denorm_fun(function(norm_fun(image_chunk)))
        if to_numpy:
            image_chunk = Backend.to_numpy(image_chunk, dtype=image.dtype)
        else:
            image_chunk = Backend.to_backend(image_chunk, dtype=image.dtype)

        remove_margin_slice_tuple = remove_margin_slice(
            shape, chunk_slice, chunk_slice_no_margins
        )
        image_chunk = image_chunk[remove_margin_slice_tuple]

        result[chunk_slice_no_margins] = image_chunk

# Dask turned out not too work great here, HUGE overhead compared to the light approach above.
# def scatter_gather_dask(backend: Backend,
#                         function,
#                         image,
#                         chunks,
#                         margins=None):
#     boundary=None
#     trim=True
#     align_arrays=True
#
#     image_d = from_array(image, chunks=chunks, asarray=False)
#
#     def function_numpy(_image):
#         print(_image.shape)
#         return backend.to_numpy(function(_image))
#
#     #func, *args, depth=None, boundary=None, trim=True, align_arrays=True, **kwargs
#     computation= map_overlap(function_numpy,
#                 image_d,
#                 depth=margins,
#                 boundary=boundary,
#                 trim=trim,
#                 align_arrays=align_arrays,
#                 dtype=image.dtype
#                 )
#
#     #computation.visualize(filename='transpose.png')
#     result = computation.compute()
#
#     return result



import itertools
from typing import List, Tuple, Any

import numpy

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.numpy_backend import NumpyBackend


def axis_aligned_pattern_correction(image,
                                    axis_combinations: List[Tuple[int]] = None,
                                    quantile: float = 0.5,
                                    sigma: float = 0,
                                    decimation: int = 8,
                                    robust_statistics: bool = True,
                                    in_place: bool = True,
                                    internal_dtype=None
                                    ):
    """
    Axis aligned pattern correction

    Corrects fixed, axis aligned, offset patterns along any combination of axis.
    Given a list of tuples of axis that defines axis-aligned volumes, intensity fluctuations
    of these volumes are stabilised.
    For example, this this class you can suppress intensity fluctuationa over time,
    suppress fixed offsets per pixel over time, suppress intensity fluctuations per row, per column. etc...


    Parameters
    ----------
    image : image to correct
    axis_combinations : List of tuples of axis in the order of correction
    percentile : percentile value used for stabilisation
    sigma : sigma used for Gaussian filtering the computed percentile values.
    in_place : True if the input image may be modified in-place.
    internal_dtype : internal dtype for computation
    """
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    if internal_dtype is None:
        internal_dtype = image.dtype

    if type(Backend.current()) is NumpyBackend:
        internal_dtype = numpy.float32

    original_dtype = image.dtype

    image = Backend.to_backend(image, dtype=internal_dtype, force_copy=not in_place)

    axis_combinations = (
        _all_axis_combinations(image.ndim)
        if axis_combinations is None
        else axis_combinations
    )

    for axis_combination in axis_combinations:
        print(f"Supressing variations across hyperplane: {axis_combination}")

        overall_value = xp.percentile(
            image.ravel()[::decimation], q=100 * quantile, keepdims=True
        )

        if robust_statistics:
            value = xp.percentile(
                image, q=100 * quantile, axis=axis_combination, keepdims=True
            )
            value = value.astype(dtype=internal_dtype, copy=False)
        else:
            value = xp.mean(
                image, axis=axis_combination, keepdims=True, dtype=internal_dtype
            )

        if sigma > 0:
            value -= sp.ndimage.filters.gaussian_filter(value, sigma=sigma)

        image += overall_value
        image -= value

    image = image.astype(original_dtype, copy=False)

    return image


def _axis_combinations(ndim: int, n: int) -> List[Tuple[Any, ...]]:
    return list(itertools.combinations(range(ndim), n))


def _all_axis_combinations(ndim: int):
    axis_combinations = []
    for dim in range(1, ndim):
        combinations = _axis_combinations(ndim, dim)
        axis_combinations.extend(combinations)
    return axis_combinations



from arbol import asection
from skimage.color import gray2rgba
from skimage.data import camera

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.color.border import add_border
from dexp.processing.color.crop_resize_pad import crop_resize_pad_color_image


def demo_crop_resize_pad_numpy():
    with NumpyBackend():
        demo_crop_resize_pad()


def demo_crop_resize_pad_cupy():
    try:
        with CupyBackend():
            demo_crop_resize_pad()
        return True
    except (ModuleNotFoundError, NotImplementedError):
        print("Cupy module not found! ignored!")
        return False


def demo_crop_resize_pad(display=True):
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    with asection("generate data"):
        image = Backend.to_backend(gray2rgba(camera()))

    # def crop_resize_pad_color_image(image,
    #                                 crop: Tuple[Tuple[int, int], ...] = None,
    #                                 resize: Tuple[int, ...] = None,
    #                                 resize_order: int = 3,
    #                                 resize_mode: str = 'constant',
    #                                 pad: Tuple[Tuple[int, int], ...] = None,
    #                                 pad_color: Tuple[float, float, float, float] = (0, 0, 0, 0),
    #                                 rgba_value_max: float = 255

    with asection("crop_resize_pad ..."):
        crop_resize_pad_image = crop_resize_pad_color_image(image,
                                                            crop=3,
                                                            resize=(500, 400),
                                                            pad_width=(3, 7)
                                                            )

        assert crop_resize_pad_image.shape == (500+10, 400+10, 4)


    if display:
        from napari import Viewer, gui_qt
        with gui_qt():
            def _c(array):
                return Backend.to_numpy(array)

            viewer = Viewer()
            viewer.add_image(_c(image), name='image', rgb=True)
            viewer.add_image(_c(crop_resize_pad_image), name='crop_resize_pad_image', rgb=True)
            viewer.grid.enabled = True


if __name__ == "__main__":
    if not demo_crop_resize_pad_cupy():
        demo_crop_resize_pad_numpy()



import click

from dexp.cli.dexp_main import _default_workers_backend
from dexp.video.crop_resize_pad import crop_resize_pad_image_sequence


@click.command()
@click.argument('input_path', type=str)
@click.option('--output_path', '-o', type=str, default=None, help='Output folder for resized frames.')
@click.option('--crop', '-c', type=str, default=None,
              help='Crop image by removing a given number of pixels/voxels per axis.'
                   ' For example: 10,20*10,20 crops 10 pixels on the left for axis 0, 20 pixels'
                   ' from the right of axis #0, and the same for axis 2.',
              show_default=True)
@click.option('--resize', '-r', type=str, default=None,
              help='After cropping, the image is resized to the given shape.'
                   ' If any entry in the tuple is -1 then that position in the shape'
                   ' is automatically determined based on the existing shape to preserve aspect ratio.',
              show_default=True)
@click.option('--resizeorder', '-ro', type=int, default=3,
              help='The order of the spline interpolation, default is 3. The order has to be in the range 0-5.',
              show_default=True)
@click.option('--resizemode', '-rm', type=str, default='constant',
              help='Optional The mode parameter determines how the input array'
                   ' is extended beyond its boundaries. Can be: ‘reflect’, ‘constant’, ‘nearest’, ‘mirror’, ‘wrap’.',
              show_default=True)
@click.option('--padwidth', '-pw', type=str, default=None,
              help='After cropping and resizing, padding is performed.'
                   ' The provided tuple is interpreted similarly to cropping.'
                   ' For example, 5,5*5,5 pads all sides with 5 pixels.',
              show_default=True)
@click.option('--padmode', '-pm', type=str, default='constant',
              help='Padding mode. See numpy.pad for the available modes',
              show_default=True)
@click.option('--padcolor', '-pc', type=str, default='0,0,0,0',
              help='Padding color as tuple of normalised floats:  (R,G,B,A). Default is transparent black.',
              show_default=True)
@click.option('--overwrite', '-w', is_flag=True,
              help='Force overwrite of output images.',
              show_default=True)
@click.option('--workers', '-k', type=int, default=-1,
              help='Number of worker threads to spawn, set to -1 for maximum number of workers',
              show_default=True)  #
@click.option('--workersbackend', '-wkb', type=str, default=_default_workers_backend,
              help='What backend to spawn workers with, can be ‘loky’ (multi-process) or ‘threading’ (multi-thread) ',
              show_default=True)  #
@click.option('--device', '-d', type=int, default=0,
              help='Sets the CUDA devices id, e.g. 0,1,2',
              show_default=True)  #
def resize(input_path,
           output_path,
           crop,
           resize,
           resizeorder,
           resizemode,
           padwidth,
           padmode,
           padcolor,
           overwrite,
           workers,
           workersbackend,
           device):
    """ Resizes video
    """

    # Default output path:
    if output_path is None:
        output_path = input_path + '_resized'
    elif output_path.startswith('_'):
        output_path = input_path + output_path

    if crop is not None:
        crop = tuple(tuple(int(v) for v in xy.split(',')) for xy in crop.split('*'))

    if resize is not None and ',' in resize:
        resize = tuple(int(v) for v in resize.split(','))

    if padwidth is not None:
        padwidth = tuple(tuple(int(v) for v in xy.split(',')) for xy in padwidth.split('*'))

    # Parse color:
    if padcolor is not None and ',' in padcolor:
        padcolor = tuple(float(v) for v in padcolor.split(','))

    crop_resize_pad_image_sequence(input_path=input_path,
                                   output_path=output_path,
                                   crop=crop,
                                   resize=resize,
                                   resize_order=resizeorder,
                                   resize_mode=resizemode,
                                   pad_width=padwidth,
                                   pad_mode=padmode,
                                   pad_color=padcolor,
                                   overwrite=overwrite,
                                   workers=workers,
                                   workersbackend=workersbackend,
                                   device=device)



from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.filters.sobel_filter import sobel_filter
from dexp.processing.synthetic_datasets.multiview_data import generate_fusion_test_data


def test_sobel_numpy():
    with NumpyBackend():
        _sobel()


def test_sobel_cupy():
    try:
        with CupyBackend():
            _sobel()
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def _sobel(length_xy=128):
    image_gt, image_lowq, blend_a, blend_b, image1, image2 = generate_fusion_test_data(length_xy=length_xy,
                                                                                       add_noise=False)

    tenengrad_image1 = sobel_filter(image1)
    tenengrad_image2 = sobel_filter(image2)

    assert tenengrad_image1.shape == image1.shape
    assert tenengrad_image2.shape == image2.shape
    assert tenengrad_image1.dtype == image2.dtype
    assert tenengrad_image2.dtype == image2.dtype

    # from napari import Viewer, gui_qt
    # with gui_qt():
    #     viewer = Viewer()
    #     viewer.add_image(image_gt, name='image_gt')
    #     viewer.add_image(image_lowq, name='image_lowq')
    #     viewer.add_image(blend_a, name='blend_a')
    #     viewer.add_image(blend_b, name='blend_b')
    #     viewer.add_image(image1, name='image1')
    #     viewer.add_image(image2, name='image2')
    #     viewer.add_image(tenengrad_image1, name='tenengrad_image1')
    #     viewer.add_image(tenengrad_image2, name='tenengrad_image2')



from arbol import asection

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.color.blend import blend_color_images
from dexp.processing.color.conversions import gray2rgba
from dexp.processing.synthetic_datasets.binary_blobs import binary_blobs


def demo_blend_numpy():
    with NumpyBackend():
        demo_blend()


def demo_blend_cupy():
    try:
        with CupyBackend():
            demo_blend()
        return True
    except (ModuleNotFoundError, NotImplementedError):
        print("Cupy module not found! ignored!")
        return False


def demo_blend(length_xy=512, display=True):
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    image_u = binary_blobs(length=length_xy, n_dim=2, blob_size_fraction=0.5, volume_fraction=0.3).astype(xp.float32)
    image_v = binary_blobs(length=length_xy, n_dim=2, blob_size_fraction=0.5, volume_fraction=0.3).astype(xp.float32)

    image_u = sp.ndimage.gaussian_filter(image_u, sigma=15)
    image_u = 255 * image_u / xp.max(image_u)
    image_v = sp.ndimage.gaussian_filter(image_v, sigma=15)
    image_v = 255 * image_v / xp.max(image_v)

    # convert to 8 bit:
    image_u = image_u.astype(xp.uint8)
    image_v = image_v.astype(xp.uint8)

    # convert to RGBA:
    image_u = gray2rgba(image_u)
    image_v = gray2rgba(image_v)

    # Keep only one channel out of 3:
    image_u[..., 0] = 0
    image_u[..., 3] = image_u[..., 1]
    image_v[..., 1] = 0
    image_v[..., 3] = image_v[..., 0]

    # Move to backend:
    image_u = Backend.to_backend(image_u)
    image_v = Backend.to_backend(image_v)

    # modulate alpha channel:
    # image_u[:, 0:256, 3] = 128
    # image_v[0:256, :, 3] = 128

    with asection("blend_mean"):
        blend_mean = blend_color_images(images=(image_u, image_v),
                                        alphas=(1, 1),
                                        modes=('max', 'mean'))

    with asection("blend_add"):
        blend_add = blend_color_images(images=(image_u, image_v),
                                       alphas=(1, 1),
                                       modes=('max', 'add'))

    # with asection("blend_erfadd"):
    #     blend_erfadd = blend((image_u, image_v), (1, 0.5), mode='erfadd')

    with asection("blend_max"):
        blend_max = blend_color_images(images=(image_u, image_v),
                                       alphas=(1, 1),
                                       modes=('max', 'max'))

    with asection("blend_over"):
        blend_over = blend_color_images(images=(image_u, image_v),
                                        alphas=(1, 1),
                                        modes=('max', 'over'))

    if display:
        from napari import Viewer, gui_qt
        with gui_qt():
            def _c(array):
                return Backend.to_numpy(array)

            viewer = Viewer()
            viewer.add_image(_c(image_u), name='image_u')
            viewer.add_image(_c(image_v), name='image_v')
            viewer.add_image(_c(blend_mean), name='blend_mean')
            viewer.add_image(_c(blend_add), name='blend_add')
            # viewer.add_image(_c(blend_erfadd), name='blend_erfadd')
            viewer.add_image(_c(blend_max), name='blend_max')
            viewer.add_image(_c(blend_over), name='blend_over')
            viewer.grid.enabled = True


if __name__ == "__main__":
    if not demo_blend_cupy():
        demo_blend_numpy()



from arbol import aprint, asection

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.registration.translation_nd_proj import register_translation_proj_nd
from dexp.processing.synthetic_datasets.multiview_data import generate_fusion_test_data


def demo_register_translation_3d_proj_diff_numpy():
    with NumpyBackend():
        _register_translation_3d_proj_diff()


def demo_register_translation_3d_proj_diff_cupy():
    try:
        with CupyBackend():
            _register_translation_3d_proj_diff()
    except ModuleNotFoundError:
        aprint("Cupy module not found! demo ignored")


def _register_translation_3d_proj_diff(length_xy=256, display=True):
    with asection("generate dataset"):
        image_gt, image_lowq, blend_a, blend_b, image1, image2 = generate_fusion_test_data(add_noise=False,
                                                                                           shift=(1, 5, -13),
                                                                                           volume_fraction=0.5,
                                                                                           length_xy=length_xy,
                                                                                           length_z_factor=1,
                                                                                           z_overlap=1)
        depth = image1.shape[0]
        crop = depth // 4

        image1 = image1[crop:-crop]
        image2 = image2[crop:-crop]

    with asection("register_translation_maxproj_nd"):
        model = register_translation_proj_nd(image1, image2)
        aprint(f"model: {model}")

    with asection("shift back"):
        image1_reg, image2_reg = model.apply_pair(image1, image2, pad=False)
        image1_reg_pad, image2_reg_pad = model.apply_pair(image1, image2, pad=True)

    if display:
        from napari import Viewer, gui_qt
        with gui_qt():
            def _c(array):
                return Backend.to_numpy(array)

            viewer = Viewer()
            viewer.add_image(_c(image_gt), name='image_gt', visible=False)
            viewer.add_image(_c(image_lowq), name='image_lowq', visible=False)
            viewer.add_image(_c(blend_a), name='blend_a', visible=False)
            viewer.add_image(_c(blend_b), name='blend_b', visible=False)
            viewer.add_image(_c(image1), name='image1', colormap='bop orange', blending='additive', visible=False)
            viewer.add_image(_c(image2), name='image2', colormap='bop blue', blending='additive', visible=False)
            viewer.add_image(_c(image1_reg), name='image1_reg', colormap='bop orange', blending='additive', visible=False)
            viewer.add_image(_c(image2_reg), name='image2_reg', colormap='bop blue', blending='additive', visible=False)
            viewer.add_image(_c(image1_reg_pad), name='image1_reg_pad', colormap='bop orange', blending='additive')
            viewer.add_image(_c(image2_reg_pad), name='image2_reg_pad', colormap='bop blue', blending='additive')

    return image1, image2, image2_reg, model


if __name__ == "__main__":
    demo_register_translation_3d_proj_diff_cupy()
    demo_register_translation_3d_proj_diff_numpy()



import numpy
from arbol import aprint

from dexp.optics.psf.microscope_psf import MicroscopePSF


def nikon16x08na(xy_size=17, z_size=17, dxy=0.485, dz=2, wvl=0.561, NA=0.8):
    return generate_psf(dxy=dxy,
                        dz=dz,
                        xy_size=xy_size,
                        z_size=z_size,
                        M=16,
                        NA=NA,
                        n=1.33,
                        wd=3000,
                        tl=165.0 * 1.0e+3,
                        wvl=wvl)


def olympus20x10na(xy_size=17, z_size=17, dxy=0.439, dz=1.8, wvl=0.561, NA=1.0):
    return generate_psf(dxy=dxy,
                        dz=dz,
                        xy_size=xy_size,
                        z_size=z_size,
                        M=20,
                        NA=NA,
                        n=1.33,
                        wd=2000,
                        tl=133.0 * 1.0e+3,
                        wvl=wvl)


def generate_psf(dxy, dz, xy_size, z_size, M=16, NA=0.8, n=1.33, wd=3000, tl=133.0 * 1.0e+3, wvl=561.):
    """
    Generates a 3D PSF array.

    :param dxy: voxel dimension along xy (microns)
    :param dz: voxel dimension along z (microns)
    :param xy_size: size of PSF kernel along x and y (odd integer)
    :param z_size: size of PSF kernel along z (odd integer)

    """

    psf_gen = MicroscopePSF()

    # Microscope parameters.
    psf_gen.parameters["M"] = M  # magnification
    psf_gen.parameters["NA"] = NA  # numerical aperture
    psf_gen.parameters["ni0"] = n
    psf_gen.parameters["ni"] = n
    psf_gen.parameters["ns"] = n
    psf_gen.parameters["ti0"] = wd
    psf_gen.parameters["zd0"] = tl

    lz = (z_size) * dz
    z_offset = -(lz - 2 * dz) / 2
    pz = numpy.arange(0, lz, dz)

    # gLXYZParticleScan(self, dxy, xy_size, pz, normalize = True, wvl = 0.6, zd = None, zv = 0.0):
    psf_xyz_array = psf_gen.gLXYZParticleScan(
        dxy=dxy,
        xy_size=xy_size,
        pz=pz,
        zv=z_offset,
        wvl=wvl
    )

    psf_xyz_array /= psf_xyz_array.sum()

    aprint(f"Generating PSF for parameters: {psf_gen.parameters}")

    return psf_xyz_array



from typing import Sequence, Union, Generator, Any, Tuple

import numpy

from dexp.processing.backends.backend import Backend


def blend_color_images(images: Union[Generator[Any, Any, None], Sequence[Any]],
                       modes: Union[str, Sequence[str]] = 'max',
                       alphas: Sequence[float] = None,
                       background_color: Tuple[float, float, float, float] = (0, 0, 0, 0),
                       rgba_value_max: float = 255,
                       internal_dtype=numpy.float32):
    """
    Blends multiple images together according to a blend mode

    Parameters
    ----------
    images: Sequence of images to blend.
    modes: Blending modes. Either one for all images, or one per image in the form of a sequence. Blending modes are: 'mean', 'add', 'satadd', 'max', 'alpha'
    alphas: Optional sequence of alpha values to use for blending, must be of same length as the sequence of images, optional.
    background_color: Background color as tuple of normalised floats:  (R,G,B,A). Default is transparent black.
    rgba_value_max: Max value for the pixel/voxel values.
    internal_dtype: dtype for internal computation

    Returns
    -------
    Blended image

    """

    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    # convert to list of images, and move to backend:
    images = list(Backend.to_backend(image) for image in images)

    # Check that there is at least one image in the image list:
    if len(images) == 0:
        raise ValueError(f"Blending requires at least one image!")

    # original dtype:
    original_dtype = images[0].dtype

    # Verify that all images are of the same shape:
    for image in images:
        if images[0].ndim != image.ndim or images[0].shape[:-1] != image.shape[:-1]:
            raise ValueError("All images in sequence must have the same number of dimensions and shape!")

    # verify that the images are 2D RGB(A) images:
    if images[0].ndim != 3 or not images[0].shape[2] in (3, 4):
        raise ValueError("Images must be 2D RGB(A) images!")

    # Check that the number of images and alphas match:
    if len(images) != len(alphas):
        raise ValueError(f"The number of images ({len(images)}) is not equal to the number of alpha values ({len(images)}).")

    # Create image to store result:
    result = xp.zeros_like(_ensure_rgba(images[0], rgba_value_max), dtype=internal_dtype)

    # Fill with background color:
    result[..., :] = xp.asarray(background_color, dtype=internal_dtype)

    # Blend in the images:
    for image, mode, alpha in zip(images, modes, alphas):
        # move image to backend:
        image = Backend.to_backend(image, dtype=internal_dtype)

        image = _ensure_rgba(image, rgba_value_max)

        # normalise image:
        image /= rgba_value_max
        result = _blend_function(alpha * image, result, mode)

    # scale back:
    result *= rgba_value_max

    # clip:
    result = xp.clip(result, 0, rgba_value_max)

    # convert back to original dtype (of first image)
    result = result.astype(dtype=original_dtype, copy=False)

    return result


def _ensure_rgba(image, value_max):
    xp = Backend.get_xp_module()
    # make sure that the image has an alpha channel:
    if image.shape[-1] == 3:
        pad_width = ((0, 0),) * (image.ndim - 1) + ((0, 1),)
        image = xp.pad(image, pad_width=pad_width)
        image[..., -1] = value_max
    return image


def _blend_function(image_u, image_v, mode):
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    if mode == 'mean':
        result = 0.5 * (image_u + image_v)

    elif mode == 'satadd':
        # TODO: this fails but it is unclear why:
        result = sp.special.erf(image_u + image_v)

    elif mode == 'max' or mode == 'rgbamax':
        result = xp.maximum(image_u, image_v)

    elif mode == 'min' or mode == 'rgbmin':
        result = xp.minimum(image_u, image_v)
        result[..., 3] = xp.maximum(image_u[..., 3], image_v[..., 3])

    elif image_u.shape[-1] == 4 and image_v.shape[-1] == 4:
        # See: https://en.wikipedia.org/wiki/Alpha_compositing
        # See: https://www.cairographics.org/operators/
        result = xp.zeros_like(image_u)

        src_rgb = image_u[..., 0:3]
        src_alpha = image_u[..., 3][..., xp.newaxis]
        dst_rgb = image_v[..., 0:3]
        dst_alpha = image_v[..., 3][..., xp.newaxis]

        if mode == 'clear':
            # Where the second object is drawn, the first is completely removed. Anywhere else it is left intact. The second object itself is not drawn.
            out_alpha = 0
            out_rgb = 0
        elif mode == 'source':
            # The second object is drawn as if nothing else were below. Only outside of the blue rectangle the red one is left intact.
            out_alpha = src_alpha
            out_rgb = src_rgb
        elif mode == 'over' or mode == 'alpha':
            # The image shows what you would expect if you held two semi-transparent slides on top of each other. This operator is cairo's default operator.
            out_alpha = (src_alpha + dst_alpha * (1 - src_alpha))
            out_rgb = (src_rgb * src_alpha + dst_rgb * dst_alpha * (1 - src_alpha)) / out_alpha
        elif mode == 'in':
            # The first object is removed completely, the second is only drawn where the first was.
            out_alpha = src_alpha * dst_alpha
            out_rgb = src_rgb
        elif mode == 'out':
            # The blue rectangle is drawn only where the red one wasn't. Since the red one was partially transparent, you can see a blue shadow in the overlapping area. Otherwise, the red object is completely removed.
            out_alpha = src_alpha * (1 - dst_alpha)
            out_rgb = src_rgb
        elif mode == 'atop':
            # This leaves the first object mostly intact, but mixes both objects in the overlapping area. The second object object is not drawn except there.
            # If you look closely, you will notice that the resulting color in the overlapping area is different from what the OVER operator produces. Any two operators produce different output in the overlapping area!
            out_alpha = dst_alpha
            out_rgb = src_alpha * src_rgb + dst_rgb * (1 - src_alpha)
        elif mode == 'dest':
            # Leaves the first object untouched, the second is discarded completely.
            out_alpha = dst_alpha
            out_rgb = dst_rgb
        elif mode == 'dest_over':
            # The result is similar to the OVER operator. Except that the "order" of the objects is reversed, so the second is drawn below the first.
            out_alpha = (1 - dst_alpha) * src_alpha + dst_alpha
            out_rgb = (src_alpha * src_rgb * (1 - dst_alpha) + dst_alpha * dst_rgb) / out_alpha
        elif mode == 'dest_in':
            # The blue rectangle is used to determine which part of the red one is left intact. Anything outside the overlapping area is removed.
            # This works like the IN operator, but again with the second object "below" the first.
            out_alpha = src_alpha * dst_alpha
            out_rgb = dst_rgb
        elif mode == 'dest_atop':
            # Same as the ATOP operator, but again as if the order of the drawing operations had been reversed.
            out_alpha = src_alpha
            out_rgb = src_rgb * (1 - dst_alpha) + dst_alpha * dst_rgb
        elif mode == 'xor':
            # The output of the XOR operator is the same for both bounded and unbounded source interpretations.
            out_alpha = src_alpha + dst_alpha - 2 * src_alpha * dst_alpha
            out_rgb = (src_alpha * src_rgb * (1 - dst_alpha) + dst_alpha * dst_rgb * (1 - src_alpha)) / out_alpha
        elif mode == 'add':
            # The output of the ADD operator is the same for both bounded and unbounded source interpretations.
            out_alpha = xp.minimum(1, src_alpha + dst_alpha)
            out_rgb = (src_alpha * src_rgb + dst_alpha * dst_rgb) / out_alpha
        elif mode == 'saturate':
            # The output of the ADD operator is the same for both bounded and unbounded source interpretations.
            out_alpha = xp.minimum(1, src_alpha + dst_alpha)
            out_rgb = (xp.minimum(src_alpha, 1 - dst_alpha) * src_rgb + dst_alpha * dst_rgb) / out_alpha
        else:
            raise ValueError(f"Invalid alpha blending mode {mode} or incompatible images: {image_u.shape} {image_v.shape}")

        result[..., 0:3] = out_rgb.squeeze()
        result[..., 3] = out_alpha.squeeze()
    else:
        raise ValueError(f"Invalid blending mode {mode} or incompatible images: {image_u.shape} {image_v.shape}")

    return result



from dexp.processing.backends.backend import Backend


def wiener_kernel(kernel,
                  alpha: float = 1e-3,
                  frequency_domain: bool = False,
                  dtype=None):
    """ Computes the Wiener filter for a given kernel and alpha parameter.

    Parameters
    ----------
    kernel : kernel
    alpha : alpha parameter
    frequency_domain : if True then the wiener kernel is returned in the frequency domain
    dtype : dtype for kernel

    Returns
    -------
    Wiener kernel
    """
    backend = Backend.current()
    xp = backend.get_xp_module()

    if dtype is None:
        dtype = kernel.dtype

    psf_fft = xp.fft.fftn(kernel)
    kernel_fft = psf_fft / (xp.abs(psf_fft) * xp.abs(psf_fft) + alpha)

    if frequency_domain:
        return kernel_fft
    else:
        kernel = xp.real(xp.fft.ifftn(kernel_fft))
        kernel = kernel / kernel.sum()
        kernel = kernel.astype(dtype, copy=False)
        return kernel



from typing import Union, Tuple, Callable

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.color.colormap import rgb_colormap, _normalise_colormap
from dexp.processing.color.insert import insert_color_image
from dexp.processing.color.projection_legend import depth_color_scale_legend
from dexp.processing.utils.center_of_mass import center_of_mass
from dexp.processing.utils.normalise import normalise_functions


def project_image(image,
                  axis: int = 0,
                  dir: int = -1,
                  mode: str = 'max',
                  attenuation: float = 0.05,
                  attenuation_min_density: float = 0.002,
                  attenuation_filtering: float = 4,
                  gamma: float = 1,
                  clim: Tuple[float, float] = None,
                  normalisation_quantile: float = 0.0001,
                  cmap: Union[str, Callable] = None,
                  dlim: Tuple[float, float] = None,
                  depth_stabilisation: bool = False,
                  rgb_gamma: float = 1,
                  transparency: bool = False,
                  legend_size: float = 0,
                  legend_scale: float = 1,
                  legend_title: str = 'depth (voxels)',
                  legend_title_color: Tuple[float, float, float, float] = (1, 1, 1, 1),
                  legend_position: Union[str, Tuple[int, int]] = 'bottom_left',
                  legend_alpha: float = 1,
                  internal_dtype=None):
    """
    Projects an image along a given axis given a specified method (max projection, max projection color-coded depth, ...)
    and produces a rendered RGB image. This function offers features similar to volume rendering:
    attenuation, projection direction direction, ...


    Parameters
    ----------
    image : Image to project
    axis : axis along which to project
    dir : projection direction, can be either '-1' for looking from top, or '1' for looking from bottom -- convention is that top corresponds to the positive direction of the projection axis.
    mode : projection mode, can be: 'max' for max projection, 'colormax' and 'maxcolor' for max color projections. Explanation: colormax applies first max and then colorises,
     maxcolor, first colorises each voxel then computes the max. This means that maxcolor is 3x more memory intensive (with current implementation).
    attenuation : How much to attenuate when projecting.
    attenuation_min_density : Minimal density for attenuation. The higher the darker structures at the back will be, independently of whether there is something in front.
    attenuation_filtering: standard deviation of the gaussian filter used to preprocess the image for the purpose of computing the attenuation map.
    gamma: Gamma correction to apply.
    clim : color limits for applying the colormap.
    normalisation_quantile : Quantile value for image normalisation (robust min and max for normalising the image between 0 and 1).
    cmap: Color map to use, can be a string or a cmap object
    Important: this does not affect sharpness of the final image, it only affects the sharpness of the attenuation itself.
    dlim: Depth limits. For example, a value of (0.1, 0.7) means that the colormap start at a normalised depth of 0.1, and ends at a normalised depth of 0.7, other values are clipped.
    depth_stabilisation: Uses the center of mass calculation to shift the center of the depth color map to teh center of mass of the image content.
    rgb_gamma: Gamma correction applied to the resulting RGB image.
    transparency: If True, the best attempt is made to use alpha transparency in the resulting image. Not available in all modes. RGB output is not 'alpha pre-multiplied'.
    legend_size: Multiplicative factor to control size of legend. If 0, no legend is generated.
    legend_scale: Float that gives the scale in some unit of each voxel (along the projection direction). Only in color projection modes.
    legend_title: title for the color-coded depth legend.
    legend_title_color: Legend title color as a tuple of normalised floats: (R, G, B, A)  (values between 0 and 1).
    legend_position: Position of the legend in pixels in natural order: (x, y). Can also be a string: 'bottom_left', 'bottom_right', 'top_left', 'top_right'.
    legend_alpha: Transparency for legend (1 means opaque, 0 means completely transparent)
    internal_dtype : dtype for internal computation

    Returns
    -------
    Projected image of shape (..., 3 or 4), legend image of shape (..., 3 or 4)

    """

    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    if internal_dtype is None:
        internal_dtype = xp.float16

    if type(Backend.current()) is NumpyBackend:
        internal_dtype = xp.float32

    # Move image to current backend:
    image = Backend.to_backend(image, dtype=internal_dtype)

    # Set default cmap:
    if cmap is None:
        if mode == 'max':
            cmap = 'viridis'
        elif mode == 'maxcolor':
            cmap = 'rainbow'
        elif mode == 'colormax':
            cmap = 'rainbow'

    # Normalise color map:
    cmap = _normalise_colormap(cmap)

    # Normalise depth limits:
    dlim = (0, 1) if dlim is None else dlim

    # Normalise image:
    norm_fun, _ = normalise_functions(image, quantile=normalisation_quantile, minmax=clim, clip=True)
    image = norm_fun(image)

    # Apply gamma:
    if gamma != 1:
        image **= gamma

    # # flip image:
    # if dir == -1 or dir == +1:
    #     if dir > 0:
    #         image = xp.flip(image, axis=axis).copy()
    # else:
    #     raise ValueError(f"Invalid direction: {dir}, must be '-1' or '+1' ")

    if attenuation != 0:

        if attenuation_filtering > 0:
            image_for_attenuation = sp.ndimage.gaussian_filter(image, sigma=attenuation_filtering)
        else:
            image_for_attenuation = image

        cum_density = _inplace_cumsum(attenuation_min_density + (1 - attenuation_min_density) * image_for_attenuation, axis=axis, dir=-dir)

        image *= xp.exp(-attenuation * cum_density)

    # Perform projection
    if mode == 'max':
        # max projection:
        projection = xp.max(image, axis=axis)

        # apply color map:
        projection = rgb_colormap(projection, cmap=cmap, bytes=False)

    elif mode == 'maxcolor':

        # compute a depth map of same shape as the image:
        depth = image.shape[axis]
        depth_values = xp.linspace(0, 1, num=depth)
        if dlim is not None:
            depth_values = _apply_depth_limits(depth_values, dlim)
        depth_values = xp.expand_dims(depth_values, axis=tuple(range(image.ndim - 1)))
        depth_values = xp.moveaxis(depth_values, -1, axis)

        # Apply the colormap:
        color_ramp = rgb_colormap(depth_values, cmap=cmap, bytes=False)

        # Multiply with image leveraging broadcasting:
        color_image = color_ramp * image[..., xp.newaxis]

        # Projecting:
        projection = xp.max(color_image, axis=axis)

    elif mode == 'colormax':

        # argmax:
        image_for_argmax = image  # xp.flip(image, axis=axis).copy() if dir < 0 else image
        indices = xp.argmax(image_for_argmax, axis=axis)

        # Equivalent to: values = xp.max(image, axis=axis) but faster because we reuse the argmax result.
        expanded_indices = xp.expand_dims(indices, axis=axis)
        values = xp.take_along_axis(image_for_argmax, expanded_indices, axis=axis)
        values = xp.squeeze(values)

        # apply color map, this is just the chroma-coding of depth
        norm_factor = xp.array(1.0 / float(image.shape[axis] - 1)).astype(internal_dtype, copy=False)
        normalised_depth = norm_factor * indices

        # Crude depth stabilisation, not recommended:
        if depth_stabilisation:
            com = center_of_mass(image)
            delta = (com[axis] / image.shape[axis]) - 0.5
            normalised_depth -= delta

        if dlim is not None:
            normalised_depth = _apply_depth_limits(normalised_depth, dlim)
        projection = rgb_colormap(normalised_depth, cmap=cmap, bytes=False)

        # Next we multiply the chroma-code with the intensity of the corresponding voxel:
        if transparency:
            # we set the alpha channel:
            projection[..., 3] *= values
        else:
            projection[..., 0:3] *= values[..., xp.newaxis]

    else:
        raise ValueError(f"Invalid projection mode: {mode}")

    if rgb_gamma != 1.0:
        if transparency:
            projection[..., 3] **= rgb_gamma
        else:
            projection[..., 0:3] **= rgb_gamma

    projection *= 255
    projection = xp.clip(projection, 0, 255)
    projection = projection.astype(xp.uint8, copy=False)

    if 'color' in mode and legend_size != 0:
        depth = image.shape[axis] * (abs(dlim[1] - dlim[0])) * legend_scale
        d_min, dmax = dlim
        start = 0
        end = depth * (dmax - d_min)
        legend = depth_color_scale_legend(cmap=cmap,
                                          start=start,
                                          end=end,
                                          flip=False,
                                          title=legend_title,
                                          color=legend_title_color,
                                          size=legend_size)

        pad_length = int(legend_size * 10)
        pad_width = ((pad_length, pad_length), (6 * pad_length, 6 * pad_length), (0, 0))
        legend = xp.pad(legend, pad_width=pad_width)

        projection = insert_color_image(projection,
                                        legend,
                                        translation=legend_position,
                                        mode='over',
                                        alpha=legend_alpha)

    return projection


def _apply_depth_limits(depth_map,
                        dlim: Tuple[float, float]):
    xp = Backend.get_xp_module()

    min_v, max_v = dlim
    depth_map = xp.clip(depth_map, min_v, max_v)
    depth_map -= depth_map.min()
    depth_map /= depth_map.max()

    return depth_map


def _inplace_cumsum(image, axis, dir):
    xp = Backend.get_xp_module()

    length = image.shape[axis]

    accumulator = xp.zeros_like(xp.take(image, 0, axis=axis))
    image_with_moved_axis = xp.moveaxis(image, axis, 0)

    positions = list(range(length))
    if dir < 0:
        positions.reverse()

    for i in positions:
        # take slice:
        _slice = xp.take(image, i, axis=axis)

        # accumulate:
        accumulator += _slice

        # place sum into array:
        image_with_moved_axis[i] = accumulator

    return image
